"""
A requests-based HTTP client for interacting with containers that have
forwarded ports.
"""
import time

import hyperlink

import requests


def _path_segments(url, path_str):
    # Absolute path
    if path_str.startswith('/'):
        return path_str[1:].split('/')
    # Relative path
    return url.child(*path_str.split('/')).path


class ContainerHttpClient:
    """
    HTTP client for a specific container.

    In most cases, these should be obtained from
    :meth:`.ContainerDefinition.http_client` instead of being instantiated
    directly.
    """

    URL_DEFAULTS = {'scheme': 'http'}

    def __init__(self, host, port, url_defaults=None, session=None):
        """
        :param host:
            The address for the host to connect to.
        :param port:
            The port for the host to connect to.
        :param dict url_defaults:
            Parameters to default to in the generated URLs, see
            `~hyperlink.URL`.
        :param session:
            A Requests' Session object (or something like it).
        """
        if session is None:
            session = requests.Session()
        self._session = session

        _url_defaults = self.URL_DEFAULTS.copy()
        if url_defaults is not None:
            _url_defaults.update(url_defaults)
        self._base_url = hyperlink.URL(
            host=host, port=int(port), **_url_defaults)

    def __enter__(self):
        return self

    def close(self):
        """
        Closes the underlying Session object.
        """
        self._session.close()

    def __exit__(self, *args):
        self.close()

    @classmethod
    def for_container(cls, container, container_port=None):
        """
        :param container:
            The container to make requests against.
        :param container_port:
            The container port to make requests against. If ``None``, the first
            container port is used.
        :returns:
            A ContainerClient object configured to make requests to the
            container.
        """
        if container_port is not None:
            host, port = container.get_host_port(container_port)
        else:
            host, port = container.get_first_host_port()

        return cls(host, port)

    def _url(self, path, kwargs):
        kwargs = kwargs if kwargs is not None else {}
        if path is not None:
            kwargs['path'] = _path_segments(self._base_url, path)
        return self._base_url.replace(**kwargs).to_text()

    def request(self, method, path=None, url_kwargs=None, **kwargs):
        """
        Make a request against a container.

        :param method:
            The HTTP method to use.
        :param list path:
            The HTTP path (either absolute or relative).
        :param dict url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param kwargs:
            Any other parameters to pass to Requests.
        """
        return self._session.request(
            method, self._url(path, url_kwargs), **kwargs)

    def get(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends a GET request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param **kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.get(self._url(path, url_kwargs), **kwargs)

    def options(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends an OPTIONS request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param **kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.options(self._url(path, url_kwargs), **kwargs)

    def head(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends a HEAD request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param **kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.head(self._url(path, url_kwargs), **kwargs)

    def post(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends a POST request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param **kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.post(self._url(path, url_kwargs), **kwargs)

    def put(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends a PUT request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param **kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.put(self._url(path, url_kwargs), **kwargs)

    def patch(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends a PUT request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param **kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.patch(self._url(path, url_kwargs), **kwargs)

    def delete(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends a PUT request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param **kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.delete(self._url(path, url_kwargs), **kwargs)


def wait_for_response(client, timeout, path='/', expected_status_code=None):
    """
    Try make a GET request with an HTTP client against a certain path and
    return once any response has been received, ignoring any errors.

    :param ContainerHttpClient client:
        The HTTP client to use to connect to the container.
    :param timeout:
        Timeout value in seconds.
    :param path:
        HTTP path to request.
    :param int expected_status_code:
        If set, wait until a response with this status code is received. If not
        set, the status code will not be checked.
    :raises TimeoutError:
        If a request fails to be made within the timeout period.
    """
    # We want time.monotonic on Pythons that have it, otherwise time.time will
    # have to do.
    get_time = getattr(time, 'monotonic', time.time)

    deadline = get_time() + timeout
    while True:
        try:
            # Don't care what the response is, as long as we get one
            time_left = deadline - get_time()
            response = client.get(
                path, timeout=max(time_left, 0.001), allow_redirects=False)

            if (expected_status_code is None
                    or response.status_code == expected_status_code):
                return
        except requests.exceptions.Timeout:
            # Requests timed out, our time must be up
            break
        except Exception:
            # Ignore other exceptions
            pass

        if get_time() >= deadline:
            break
        time.sleep(0.1)

    raise TimeoutError('Timeout waiting for HTTP response.')



"""
Some (optional) utilities for use with pytest.

While Seaworthy doesn't require pytest, we find it useful in downstream
container tests we write with Seaworthy. This module contains various bits and
pieces to make Seaworthy work better with pytest.
"""
from .checks import dockertest
from .fixtures import docker_helper

__all__ = ['docker_helper', 'dockertest']



"""
PostgreSQL container definition.
"""

from seaworthy.definitions import ContainerDefinition
from seaworthy.utils import output_lines


class PostgreSQLContainer(ContainerDefinition):
    """
    PostgreSQL container definition.

    .. todo::
       Write more docs.
    """

    DEFAULT_NAME = 'postgresql'
    DEFAULT_IMAGE = 'postgres:alpine'
    # The postgres image starts up PostgreSQL twice--the first time to set up
    # the database and user, and the second to actually run the thing.
    DEFAULT_WAIT_PATTERNS = (
        r'database system is ready to accept connections',
        r'database system is ready to accept connections',)

    DEFAULT_DATABASE = 'database'
    DEFAULT_USER = 'user'
    DEFAULT_PASSWORD = 'password'

    def __init__(self,
                 name=DEFAULT_NAME,
                 image=DEFAULT_IMAGE,
                 wait_patterns=DEFAULT_WAIT_PATTERNS,
                 database=DEFAULT_DATABASE,
                 user=DEFAULT_USER,
                 password=DEFAULT_PASSWORD,
                 **kwargs):
        """
        :param database: the name of a database to create at startup
        :param user: the name of a user to create at startup
        :param password: the password for the user
        """
        super().__init__(name, image, wait_patterns, **kwargs)

        self.database = database
        self.user = user
        self.password = password

    def base_kwargs(self):
        """
        Add a ``tmpfs`` entry for ``/var/lib/postgresql/data`` to avoid
        unnecessary disk I/O and ``environment`` entries for the configured db
        and user creds.
        """
        return {
            'environment': {
                'POSTGRES_DB': self.database,
                'POSTGRES_USER': self.user,
                'POSTGRES_PASSWORD': self.password,
            },
            'tmpfs': {'/var/lib/postgresql/data': 'uid=70,gid=70'},
        }

    def exec_pg_success(self, cmd):
        """
        Execute a command inside a running container as the postgres user,
        asserting success.
        """
        result = self.inner().exec_run(cmd, user='postgres')
        assert result.exit_code == 0, result.output.decode('utf-8')
        return result

    def clean(self):
        """
        Remove all data by dropping and recreating the configured database.

        .. note::

            Only the configured database is removed. Any other databases
            remain untouched.
        """
        self.exec_pg_success(['dropdb', '-U', self.user, self.database])
        self.exec_pg_success(['createdb', '-U', self.user, self.database])

    def exec_psql(self, command, psql_opts=['-qtA']):
        """
        Execute a ``psql`` command inside a running container. By default the
        container's database is connected to.

        :param command: the command to run (passed to ``-c``)
        :param psql_opts: a list of extra options to pass to ``psql``
        :returns: a tuple of the command exit code and output
        """
        cmd = ['psql'] + psql_opts + [
            '--dbname', self.database,
            '-U', self.user,
            '-c', command,
        ]
        return self.inner().exec_run(cmd, user='postgres')

    def list_databases(self):
        """
        Runs the ``\\list`` command and returns a list of column values with
        information about all databases.
        """
        lines = output_lines(self.exec_psql('\\list'))
        return [line.split('|') for line in lines]

    def list_tables(self):
        """
        Runs the ``\\dt`` command and returns a list of column values with
        information about all tables in the database.
        """
        lines = output_lines(self.exec_psql('\\dt'))
        return [line.split('|') for line in lines]

    def list_users(self):
        """
        Runs the ``\\du`` command and returns a list of column values with
        information about all user roles.
        """
        lines = output_lines(self.exec_psql('\\du'))
        return [line.split('|') for line in lines]

    def database_url(self):
        """
        Returns a "database URL" for use with DJ-Database-URL and similar
        libraries.
        """
        return 'postgres://{}:{}@{}/{}'.format(
            self.user, self.password, self.name, self.database)



"""
A number of pytest fixtures or factories for fixtures.
"""

import os

import pytest

from seaworthy.definitions import ContainerDefinition, _DefinitionBase
from seaworthy.helpers import DockerHelper


def docker_helper_fixture(name='docker_helper', scope='module', **kwargs):
    """
    Create a fixture for :class:`~seaworthy.DockerHelper`.

    This can be used to create a fixture with a different name to the default.
    It can also be used to override the scope of the default fixture::

        docker_helper = docker_helper_fixture(scope='class')

    :param name: The name of the fixture.
    :param scope: The scope of the fixture.
    :param kwargs:
        Keyword arguments to pass to the :class:`~seaworthy.DockerHelper`
        constructor.
    """
    @pytest.fixture(name=name, scope=scope)
    def fixture():
        namespace = kwargs.pop('namespace', 'test')
        if 'PYTEST_XDIST_WORKER' in os.environ:  # pragma: no cover
            namespace = '{}_{}'.format(
                namespace, os.environ['PYTEST_XDIST_WORKER'])
        docker_helper = DockerHelper(namespace=namespace, **kwargs)
        yield docker_helper
        docker_helper.teardown()
    return fixture


#: Default fixture for :class:`~seaworthy.DockerHelper`. Has module scope.
docker_helper = docker_helper_fixture()


def image_fetch_fixture(image, name, scope='module'):
    """
    Create a fixture to fetch an image.
    """
    @pytest.fixture(name=name, scope=scope)
    def fixture(docker_helper):
        return docker_helper.images.fetch(image)
    return fixture


def resource_fixture(definition, name, scope='function', dependencies=()):
    """
    Create a fixture for a resource.

    .. note:: This function returns a fixture function. It is important to keep
        a reference to the returned function within the scope of the tests that
        use the fixture.

    .. code-block:: python

        fixture = resource_fixture(PostgreSQLContainer(), 'postgresql')

        def test_container(postgresql):
            \"""Test something about the PostgreSQL container...\"""

    :param definition:
        A resource definition, one of those defined in the
        :mod:`seaworthy.definitions` module.
    :param name: The fixture name.
    :param scope: The scope of the fixture.
    :param dependencies:
        A sequence of names of other pytest fixtures that this fixture depends
        on. These fixtures will be requested from pytest and so will be setup,
        but nothing is done with the actual fixture values.

    :returns: The fixture function.
    """
    @pytest.fixture(name=name, scope=scope)
    def fixture(request, docker_helper):
        for dependency in dependencies:
            request.getfixturevalue(dependency)

        definition.setup(helper=docker_helper)
        yield definition
        definition.teardown()

    return fixture


def _definition_fixture(self, name, scope='function', dependencies=()):
    """
    Create a pytest fixture for the resource. See :func:`.resource_fixture`.

    .. note:: This method returns a fixture function. It is important to
        keep a reference to the returned function within the scope of the
        tests that use the fixture.

    .. note:: This method is only available if pytest is used.

    :param name: The fixture name.
    :param scope: The scope of the fixture.
    :param dependencies:
        A sequence of names of other pytest fixtures that this fixture
        depends on. These fixtures will be requested from pytest and so
        will be setup, but nothing is done with the actual fixture values.
    """
    return resource_fixture(self, name, scope, dependencies)


_DefinitionBase.pytest_fixture = _definition_fixture


def _clean_container_fixture(name, raw_name):
    @pytest.fixture(name=name)
    def clean_fixture(request):
        container = request.getfixturevalue(raw_name)
        if 'clean_{}'.format(name) in request.keywords:
            container.clean()
        return container

    return clean_fixture


def clean_container_fixtures(container, name, scope='class', dependencies=()):
    """
    Creates a fixture for a container that can be "cleaned". When a code block
    is marked with ``@pytest.mark.clean_<fixture name>`` then the ``clean``
    method will be called on the container object before it is passed as an
    argument to the test function.

    .. note:: This function returns two fixture functions. It is important to
        keep references to the returned functions within the scope of the tests
        that use the fixtures.

    .. code-block:: python

        f1, f2 = clean_container_fixtures(PostgreSQLContainer(), 'postgresql')

        class TestPostgresqlContainer
            @pytest.mark.clean_postgresql
            def test_clean_container(self, web_container, postgresql):
                \"""
                Test something about the container that requires it to have a
                clean state (e.g. database table creation).
                \"""

            def test_dirty_container(self, web_container, postgresql):
                \"""
                Test something about the container that doesn't require it to
                have a clean state (e.g. testing something about a dependent
                container).
                \"""

    :param container:
        A "container" object that is a subclass of
        :class:`.ContainerDefinition`.
    :param name:
        The fixture name.
    :param scope:
        The scope of the fixture.
    :param dependencies:
        A sequence of names of other pytest fixtures that this fixture depends
        on. These fixtures will be requested from pytest and so will be setup,
        but nothing is done with the actual fixture values.

    :returns:
        A tuple of two fixture functions.
    """
    raw_name = 'raw_{}'.format(name)
    return (resource_fixture(container, raw_name, scope, dependencies),
            _clean_container_fixture(name, raw_name))


def _definition_clean_fixtures(self, name, scope='function', dependencies=()):
    """
    Creates a pytest fixture for a container that can be "cleaned". See
    :func:`.clean_container_fixtures`.

    .. note:: This method returns two fixture functions. It is important to
        keep references to the returned functions within the scope of the
        tests that use the fixtures.

    .. note:: This method is only available if pytest is used.

    :param name: The fixture name.
    :param scope: The scope of the fixture.
    :param dependencies:
        A sequence of names of other pytest fixtures that this fixture
        depends on. These fixtures will be requested from pytest and so
        will be setup, but nothing is done with the actual fixture values.
    """
    return clean_container_fixtures(self, name, scope, dependencies)


ContainerDefinition.pytest_clean_fixtures = _definition_clean_fixtures


__all__ = ['clean_container_fixtures', 'docker_helper',
           'docker_helper_fixture', 'image_fetch_fixture', 'resource_fixture']



import unittest

import requests.exceptions

import responses

from seaworthy.checks import docker_client, dockertest
from seaworthy.client import ContainerHttpClient, wait_for_response
from seaworthy.definitions import ContainerDefinition
from seaworthy.helpers import DockerHelper, fetch_images


# Small (<4MB) image that echoes HTTP requests and runs without configuration
IMG = 'jmalloc/echo-server'


@dockertest()
def setUpModule():  # noqa: N802 (The camelCase is mandated by unittest.)
    with docker_client() as client:
        fetch_images(client, [IMG])


class DummySession:
    def __init__(self):
        self.requests = []
        self.was_closed = False

    def request(self, *args, **kwargs):
        self.requests.append((args, kwargs))

    def close(self):
        self.was_closed = True

    def check_was_closed(self):
        was_closed, self.was_closed = self.was_closed, False
        return was_closed


class TestContainerHttpClient(unittest.TestCase):
    def make_helper(self):
        dh = DockerHelper()
        self.addCleanup(dh.teardown)
        return dh.containers

    @responses.activate
    def test_defaults(self):
        """
        When the container client is configured with a host address and port,
        requests are made to that address and port.
        """
        client = ContainerHttpClient('127.0.0.1', '12345')

        responses.add(responses.GET, 'http://127.0.0.1:12345/', status=200)
        response = client.request('GET')

        self.assertEqual(response.status_code, 200)

        [call] = responses.calls
        self.assertEqual(call.request.url, 'http://127.0.0.1:12345/')

    @responses.activate
    def test_url_defaults(self):
        """
        When the container client is configured with a host address and port,
        and some URL defaults are set, requests are made to that address and
        port with the expected URL.
        """
        client = ContainerHttpClient('127.0.0.1', '12345', url_defaults={
            'scheme': 'https',
            'fragment': 'test',
        })

        responses.add(responses.GET, 'https://127.0.0.1:12345/baz', status=200)
        response = client.request('GET', '/baz', url_kwargs={
            'query': (('foo', 'bar'),),
        })

        self.assertEqual(response.status_code, 200)

        [call] = responses.calls
        self.assertEqual(
            call.request.url, 'https://127.0.0.1:12345/baz?foo=bar#test')

    @responses.activate
    def test_paths(self):
        """
        The path is appended to the URL correctly with various leading or
        trailing ``/`` characters.
        """
        client = ContainerHttpClient('127.0.0.1', '12345')

        # Root path
        responses.add(responses.GET, 'http://127.0.0.1:12345/', status=200)
        client.request('GET', '')  # Requests adds a trailing /
        client.request('GET', '/')

        self.assertEqual(
            responses.calls[0].request.url, 'http://127.0.0.1:12345/')
        self.assertEqual(
            responses.calls[1].request.url, 'http://127.0.0.1:12345/')

        # Leading slashes are ignored
        responses.add(
            responses.GET, 'http://127.0.0.1:12345/a/b/c', status=200)
        client.request('GET', '/a/b/c')
        client.request('GET', 'a/b/c')

        self.assertEqual(
            responses.calls[2].request.url, 'http://127.0.0.1:12345/a/b/c')
        self.assertEqual(
            responses.calls[3].request.url, 'http://127.0.0.1:12345/a/b/c')

        # Trailing slashes are respected
        responses.add(
            responses.GET, 'http://127.0.0.1:12345/a/b/c/', status=200)
        client.request('GET', '/a/b/c/')

        self.assertEqual(
            responses.calls[4].request.url, 'http://127.0.0.1:12345/a/b/c/')

        # Double slashes are not ignored
        responses.add(
            responses.GET, 'http://127.0.0.1:12345//a//b', status=200)
        client.request('GET', '//a//b')

        self.assertEqual(
            responses.calls[5].request.url, 'http://127.0.0.1:12345//a//b')

    @responses.activate
    def test_relative_paths(self):
        """
        The path can be specified as a relative or absolute path.
        """
        client = ContainerHttpClient(
            '127.0.0.1', '12345', url_defaults={'path': ['foo']})

        responses.add(
            responses.GET, 'http://127.0.0.1:12345/foo/bar/baz', status=200)
        client.request('GET', 'bar/baz')

        self.assertEqual(responses.calls[0].request.url,
                         'http://127.0.0.1:12345/foo/bar/baz')

        responses.add(
            responses.GET, 'http://127.0.0.1:12345/foobar', status=200)
        client.request('GET', '/foobar')

        self.assertEqual(responses.calls[1].request.url,
                         'http://127.0.0.1:12345/foobar')

    @responses.activate
    def test_methods(self):
        """
        When the HTTP method-specific methods are called, the correct request
        method is used.
        """
        client = ContainerHttpClient('127.0.0.1', '45678')

        responses.add(responses.GET, 'http://127.0.0.1:45678/', status=200)
        responses.add(
            responses.OPTIONS, 'http://127.0.0.1:45678/foo', status=201)
        responses.add(responses.HEAD, 'http://127.0.0.1:45678/bar', status=403)
        responses.add(responses.POST, 'http://127.0.0.1:45678/baz', status=404)
        responses.add(responses.PUT, 'http://127.0.0.1:45678/test', status=418)
        responses.add(
            responses.PATCH, 'http://127.0.0.1:45678/a/b/c', status=501)
        responses.add(
            responses.DELETE, 'http://127.0.0.1:45678/d/e/f', status=503)

        get_response = client.get()
        options_response = client.options('/foo')
        head_response = client.head('/bar')
        post_response = client.post('/baz')
        put_response = client.put('/test')
        patch_response = client.patch('/a/b/c')
        delete_response = client.delete('/d/e/f')

        self.assertEqual(get_response.status_code, 200)
        self.assertEqual(options_response.status_code, 201)
        self.assertEqual(head_response.status_code, 403)
        self.assertEqual(post_response.status_code, 404)
        self.assertEqual(put_response.status_code, 418)
        self.assertEqual(patch_response.status_code, 501)
        self.assertEqual(delete_response.status_code, 503)

        self.assertEqual(len(responses.calls), 7)

    def test_session(self):
        """
        When a custom session object is given, that object is used to make
        requests and is closed when ``close()`` is called.
        """
        session = DummySession()
        client = ContainerHttpClient('127.0.0.1', '12345', session=session)

        client.request('GET', '/foo')
        client.request('POST', '/bar')
        self.assertEqual(session.requests, [
            (('GET', 'http://127.0.0.1:12345/foo'), {}),
            (('POST', 'http://127.0.0.1:12345/bar'), {}),
        ])

        client.close()
        self.assertTrue(session.check_was_closed())

    def test_session_context_manager(self):
        """
        When a custom session object is given, that object is used to make
        requests and is closed when the context is exited when the container
        client is used as a context manager.
        """
        session = DummySession()
        client = ContainerHttpClient('127.0.0.1', '12345', session=session)

        with client:
            client.request('GET', '/foo')
            self.assertEqual(session.requests, [
                (('GET', 'http://127.0.0.1:12345/foo'), {}),
            ])

        self.assertTrue(session.check_was_closed())

    @dockertest()
    def test_for_container_first_port(self):
        """
        The ``for_container()`` class method returns a container client that
        connects to the container's first port when a specific port is not
        specified.
        """
        ch = self.make_helper()
        container = ContainerDefinition('first_port', IMG, create_kwargs={
            'ports': {'8080/tcp': ('127.0.0.1', None)}
        }, helper=ch)
        container.setup()
        self.addCleanup(container.teardown)

        client = ContainerHttpClient.for_container(container)
        self.addCleanup(client.close)

        response = client.request('GET', '/foo')

        self.assertEqual(response.status_code, 200)
        response_lines = response.text.splitlines()
        self.assertIn('HTTP/1.1 GET /foo', response_lines)

        addr, port = container.get_first_host_port()
        self.assertIn('Host: {}:{}'.format(addr, port), response_lines)

    @dockertest()
    def test_for_container_specific_port(self):
        """
        The ``for_container()`` class method returns a container client that
        connects to the container port specified.
        """
        ch = self.make_helper()
        container = ContainerDefinition('first_port', IMG, create_kwargs={
            'ports': {
                '8080/tcp': ('127.0.0.1', None),
                '5353/udp': ('127.0.0.1', None),
            }
        }, helper=ch)
        container.setup()
        self.addCleanup(container.teardown)

        client = ContainerHttpClient.for_container(
            container, container_port='8080')
        self.addCleanup(client.close)

        response = client.request('GET', '/foo')

        self.assertEqual(response.status_code, 200)
        response_lines = response.text.splitlines()
        self.assertIn('HTTP/1.1 GET /foo', response_lines)

        addr, port = container.get_host_port('8080')
        self.assertIn('Host: {}:{}'.format(addr, port), response_lines)


class TestWaitForResponseFunc(unittest.TestCase):
    @responses.activate
    def test_success(self):
        """
        When a request succeeds before the timeout, all is happy.
        """
        client = ContainerHttpClient('127.0.0.1', '12345')
        responses.add(responses.GET, 'http://127.0.0.1:12345/', status=200)
        # A failure here will raise an exception.
        # 100ms is long enough for a first-time success.
        wait_for_response(client, 0.1)

    @responses.activate
    def test_success_with_status_code(self):
        """
        When a request succeeds before the timeout and has the expected status
        code, all is happy.
        """
        client = ContainerHttpClient('127.0.0.1', '12345')
        responses.add(responses.GET, 'http://127.0.0.1:12345/', status=200)
        # A failure here will raise an exception.
        # 100ms is long enough for a first-time success.
        wait_for_response(client, 0.1, expected_status_code=200)

    @responses.activate
    def test_error_then_success(self):
        """
        When an exception is raised before the timeout, we retry and are happy
        with any successful request before the timeout.
        """
        client = ContainerHttpClient('127.0.0.1', '12345')
        responses.add(
            responses.GET, 'http://127.0.0.1:12345/', body=Exception('KABOOM'))
        responses.add(responses.GET, 'http://127.0.0.1:12345/', status=200)
        # A failure here will raise an exception.
        # Because responses is fast 110ms gives us time to fail, wait 100ms,
        # then succeed.
        wait_for_response(client, 0.11)

    @responses.activate
    def test_error_timeout(self):
        """
        When exceptions are raised without a successful request before the
        timeout, we time out.
        """
        client = ContainerHttpClient('127.0.0.1', '12345')
        responses.add(
            responses.GET, 'http://127.0.0.1:12345/', body=Exception('KABOOM'))
        with self.assertRaises(TimeoutError) as cm:
            # 190ms is enough time to fail, wait 100ms, fail again, wait 100ms,
            # then time out.
            wait_for_response(client, 0.19)
        self.assertEqual(
            str(cm.exception), 'Timeout waiting for HTTP response.')

    @responses.activate
    def test_unexpected_status_code_timeout(self):
        """
        When requests are received without the correct status code before the
        timeout, we time out.
        """
        client = ContainerHttpClient('127.0.0.1', '12345')
        responses.add(
            responses.GET, 'http://127.0.0.1:12345/', status=503)
        with self.assertRaises(TimeoutError) as cm:
            # 190ms is enough time to fail, wait 100ms, fail again, wait 100ms,
            # then time out.
            wait_for_response(client, 0.19, expected_status_code=200)
        self.assertEqual(
            str(cm.exception), 'Timeout waiting for HTTP response.')

    @responses.activate
    def test_timeout(self):
        """
        When we don't get a response before the timeout, we time out.

        FIXME: Because responses doesn't do timeouts, we fake it by manually
        raising the exception we expect. We really should use requests itself
        for this.
        """
        client = ContainerHttpClient('127.0.0.1', '12345')

        responses.add(
            responses.GET, 'http://127.0.0.1:12345/',
            body=requests.exceptions.Timeout())
        with self.assertRaises(TimeoutError) as cm:
            # The timeout doesn't actually matter here because we raise the
            # exception ourselves.
            wait_for_response(client, 0.1)
        self.assertEqual(
            str(cm.exception), 'Timeout waiting for HTTP response.')



"""
Some (optional) utilities for use with testtools.

While Seaworthy doesn't require testtools, we find it useful in downstream
container tests we write with Seaworthy. This module contains various bits and
pieces to make Seaworthy work better with testtools.
"""

from testtools.matchers import MatchesSetwise, MatchesStructure, Mismatch


class PsTreeMismatch(Mismatch):
    """
    Custom mismatch container for MatchesPsTree so we get somewhat more
    comprehensible failure messages.
    """

    def __init__(self, row_fields, child_count, fields_mm, children_mm):
        self.row_fields = row_fields
        self.child_count = child_count
        self.fields_mm = fields_mm
        self.children_mm = children_mm

    def describe(self):
        """
        Describe the mismatch.
        """
        rfs = ['{}={!r}'.format(k, v)
               for k, v in sorted(self.row_fields.items())]
        suffix = '' if self.child_count == 1 else 'ren'
        descriptions = ['PsTree({} with {} child{}) mismatch: ['.format(
            ', '.join(rfs), self.child_count, suffix)]
        if self.fields_mm is not None:
            for m in self.fields_mm.mismatches:
                for l in m.describe().splitlines():
                    descriptions.append('  ' + l.rstrip('\n'))
        if self.children_mm is not None:
            descriptions.append('  mismatches in children:')
            for l in self.children_mm.describe().splitlines():
                descriptions.append('    ' + l.rstrip('\n'))
        descriptions.append(']')
        return '\n'.join(descriptions)


class MatchesPsTree:
    """
    Matches a nested PsTree object in a sensible way.

    The `ruser` and `args` fields are always checked. The `pid` and `ppid`
    fields are only checked if non-None values are explicitly provided, because
    real pids are essentially arbitrary integers. The `children` field is
    always checked, but order is ignored.
    """

    def __init__(self, ruser, args, pid=None, ppid=None, children=()):
        self.row_fields = {'ruser': ruser, 'args': args}
        if pid is not None:
            self.row_fields['pid'] = pid
        if ppid is not None:
            self.row_fields['ppid'] = ppid
        self.children = children

    def __str__(self):
        rfs = ['{}={!r}'.format(k, v)
               for k, v in sorted(self.row_fields.items())]
        return '{}({}, children={})'.format(
            self.__class__.__name__, ', '.join(rfs), str(self.children))

    def match(self, value):
        """
        Return ``None`` if this matcher matches something, a
        :class:`PsTreeMismatch` otherwise.
        """
        fields_mm = MatchesStructure.byEquality(**self.row_fields).match(
            value.row)
        children_mm = MatchesSetwise(*self.children).match(value.children)
        if fields_mm is not None or children_mm is not None:
            return PsTreeMismatch(
                self.row_fields, len(self.children), fields_mm, children_mm)



import tempfile
import unittest

import docker
from docker import models

from seaworthy.checks import docker_client, dockertest
from seaworthy.helpers import (
    ContainerHelper, DockerHelper, ImageHelper, NetworkHelper, VolumeHelper,
    _parse_image_tag, fetch_images)


# We use this image to test with because it is a small (~7MB) image from
# https://github.com/docker-library/official-images that runs indefinitely with
# no configuration.
IMG = 'nginx:alpine'


@dockertest()
def setUpModule():  # noqa: N802 (The camelCase is mandated by unittest.)
    with docker_client() as client:
        fetch_images(client, [IMG])


def filter_by_name(things, prefix):
    return [t for t in things if t.name.startswith(prefix)]


# This is a private function but it's difficult to test indirectly without
# pulling images.
class TestParseImageTagFunc(unittest.TestCase):
    def test_with_tag(self):
        """An image name with a tag is parsed into a name and tag."""
        self.assertEqual(('test', 'foo'), _parse_image_tag('test:foo'))

    def test_without_tag(self):
        """An image name without a tag is parsed into a name and None."""
        self.assertEqual(('test', None), _parse_image_tag('test'))

    def test_with_tag_and_registry(self):
        """
        An image name with a tag and a registry is parsed into a name and tag.
        """
        self.assertEqual(('myregistry:5000/test', 'foo'),
                         _parse_image_tag('myregistry:5000/test:foo'))

    def test_without_tag_with_registry(self):
        """
        An image name without a tag but with a registry is parsed into a name
        and None.
        """
        self.assertEqual(('myregistry:5000/test', None),
                         _parse_image_tag('myregistry:5000/test'))


@dockertest()
class TestImageHelper(unittest.TestCase):
    def setUp(self):
        self.client = docker.client.from_env()
        self.addCleanup(self.client.api.close)

    def make_helper(self):
        return ImageHelper(self.client)

    def test_fetch(self):
        """
        We check if the image is already present and pull it if necessary.
        """
        ih = self.make_helper()

        # First, remove the image if it's already present. (We use the busybox
        # image for this test because it's the smallest I can find that is
        # likely to be reliably available.)
        try:
            self.client.images.get('busybox:latest')
        except docker.errors.ImageNotFound:  # pragma: no cover
            pass
        else:
            self.client.images.remove('busybox:latest')  # pragma: no cover

        # Pull the image, which we now know we don't have.
        with self.assertLogs('seaworthy', level='INFO') as cm:
            ih.fetch('busybox:latest')
        self.assertEqual(
            [l.getMessage() for l in cm.records],
            ["Pulling tag 'latest' for image 'busybox'..."])

        # Pull the image again, now that we know it's present.
        with self.assertLogs('seaworthy', level='DEBUG') as cm:
            ih.fetch('busybox:latest')
        logs = [l.getMessage() for l in cm.records]
        self.assertEqual(len(logs), 1)
        self.assertRegex(
            logs[0],
            r"Found image 'sha256:[a-f0-9]{64}' for tag 'busybox:latest'")


@dockertest()
class TestNetworkHelper(unittest.TestCase):
    def setUp(self):
        self.client = docker.client.from_env()
        self.addCleanup(self.client.api.close)

    def make_helper(self, namespace='test'):
        nh = NetworkHelper(self.client, namespace)
        self.addCleanup(nh._teardown)
        return nh

    def list_networks(self, *args, namespace='test', **kw):
        return filter_by_name(
            self.client.networks.list(*args, **kw), '{}_'.format(namespace))

    def test_default_lifecycle(self):
        """
        The default network can only be created once and is removed during
        teardown.
        """
        nh = self.make_helper()
        # The default network isn't created unless required
        network = nh.get_default(create=False)
        self.assertIsNone(network)

        # Create the default network
        network = nh.get_default(create=True)
        self.assertIsNotNone(network)

        # We can try to get the network lots of times and we get the same one
        # and new ones aren't created.
        nh.get_default(create=False)
        nh.get_default(create=True)
        networks = self.list_networks()
        self.assertEqual(networks, [network])

        # The default network is removed on teardown
        nh._teardown()
        network = nh.get_default(create=False)
        self.assertIsNone(network)

    def test_default_already_exists(self):
        """
        If the default network already exists when we try to create it, we
        fail.
        """
        # We use a separate NetworkHelper (with the usual cleanup) to create
        # the test network so that the DockerHelper under test will see that it
        # already exists.
        nh1 = self.make_helper()
        nh1.get_default()
        # Now for the test.
        nh2 = self.make_helper()
        with self.assertRaises(docker.errors.APIError) as cm:
            nh2.get_default()
        self.assertIn('network', str(cm.exception))
        self.assertIn('already exists', str(cm.exception))

    def test_teardown(self):
        """
        NetworkHelper._teardown() will remove any networks that were created,
        even if they no longer exist.
        """
        nh = self.make_helper()
        self.assertEqual([], self.list_networks())
        net_bridge1 = nh.create('bridge1', driver='bridge')
        net_bridge2 = nh.create('bridge2', driver='bridge')

        net_removed = nh.create('removed')
        # We remove this behind the helper's back so the helper thinks it still
        # exists at teardown time.
        net_removed.remove()
        with self.assertRaises(docker.errors.NotFound):
            net_removed.reload()

        self.assertEqual(
            set([net_bridge1, net_bridge2]),
            set(self.list_networks()))

        with self.assertLogs('seaworthy', level='WARNING') as cm:
            nh._teardown()
        self.assertEqual(sorted(l.getMessage() for l in cm.records), [
            "Network 'test_bridge1' still existed during teardown",
            "Network 'test_bridge2' still existed during teardown",
        ])
        self.assertEqual([], self.list_networks())

    def test_create(self):
        """
        We can create a network with various parameters.
        """
        nh = self.make_helper()

        net_simple = nh.create('simple')
        self.addCleanup(nh.remove, net_simple)
        self.assertEqual(net_simple.name, 'test_simple')
        self.assertEqual(net_simple.attrs['Driver'], 'bridge')
        self.assertEqual(net_simple.attrs['Internal'], False)

        net_internal = nh.create('internal', internal=True)
        self.addCleanup(nh.remove, net_internal)
        self.assertEqual(net_internal.name, 'test_internal')
        self.assertEqual(net_internal.attrs['Internal'], True)

        # Copy custom IPAM/subnet example from Docker docs:
        # https://docker-py.readthedocs.io/en/2.5.1/networks.html#docker.models.networks.NetworkCollection.create
        ipam_pool = docker.types.IPAMPool(
            subnet='192.168.52.0/24', gateway='192.168.52.254')
        ipam_config = docker.types.IPAMConfig(pool_configs=[ipam_pool])
        net_subnet = nh.create('subnet', ipam=ipam_config)
        self.addCleanup(nh.remove, net_subnet)
        self.assertEqual(net_subnet.name, 'test_subnet')
        config = net_subnet.attrs['IPAM']['Config'][0]
        self.assertEqual(config['Subnet'], '192.168.52.0/24')
        self.assertEqual(config['Gateway'], '192.168.52.254')

    def test_remove(self):
        """
        We can remove a network.
        """
        nh = self.make_helper()

        net_test = nh.create('test')
        nh.remove(net_test)
        with self.assertRaises(docker.errors.NotFound):
            net_test.reload()

    def test_custom_namespace(self):
        """
        When the helper has a custom namespace, the networks created are
        prefixed with the namespace.
        """
        nh = self.make_helper(namespace='integ')

        net = nh.create('net')
        self.addCleanup(nh.remove, net)
        self.assertEqual(net.name, 'integ_net')


@dockertest()
class TestVolumeHelper(unittest.TestCase):
    def setUp(self):
        self.client = docker.client.from_env()
        self.addCleanup(self.client.api.close)

    def make_helper(self, namespace='test'):
        vh = VolumeHelper(self.client, namespace)
        self.addCleanup(vh._teardown)
        return vh

    def list_volumes(self, *args, namespace='test', **kw):
        return filter_by_name(
            self.client.volumes.list(*args, **kw), '{}_'.format(namespace))

    def test_teardown(self):
        """
        VolumeHelper._teardown() will remove any volumes that were created,
        even if they no longer exist.
        """
        vh = self.make_helper()
        self.assertEqual([], self.list_volumes())
        vol_local1 = vh.create('local1', driver='local')
        vol_local2 = vh.create('local2', driver='local')

        vol_removed = vh.create('removed')
        # We remove this behind the helper's back so the helper thinks it still
        # exists at teardown time.
        vol_removed.remove()
        with self.assertRaises(docker.errors.NotFound):
            vol_removed.reload()

        self.assertEqual(
            set([vol_local1, vol_local2]),
            set(self.list_volumes()))

        with self.assertLogs('seaworthy', level='WARNING') as cm:
            vh._teardown()
        self.assertEqual(sorted(l.getMessage() for l in cm.records), [
            "Volume 'test_local1' still existed during teardown",
            "Volume 'test_local2' still existed during teardown",
        ])
        self.assertEqual([], self.list_volumes())

    def test_create(self):
        """
        We can create a volume with various parameters.
        """
        vh = self.make_helper()

        vol_simple = vh.create('simple')
        self.addCleanup(vh.remove, vol_simple)
        self.assertEqual(vol_simple.name, 'test_simple')
        self.assertEqual(vol_simple.attrs['Driver'], 'local')

        vol_labels = vh.create('labels', labels={'foo': 'bar'})
        self.addCleanup(vh.remove, vol_labels)
        self.assertEqual(vol_labels.name, 'test_labels')
        self.assertEqual(vol_labels.attrs['Labels'], {'foo': 'bar'})

        # Copy tmpfs example from Docker docs:
        # https://docs.docker.com/engine/reference/commandline/volume_create/#driver-specific-options
        # This won't work on Windows
        driver_opts = {
            'type': 'tmpfs', 'device': 'tmpfs', 'o': 'size=100m,uid=1000'}
        vol_opts = vh.create('opts', driver='local', driver_opts=driver_opts)
        self.addCleanup(vh.remove, vol_opts)
        self.assertEqual(vol_opts.name, 'test_opts')
        self.assertEqual(vol_opts.attrs['Options'], driver_opts)

    def test_remove(self):
        """
        We can remove a volume.
        """
        vh = self.make_helper()

        vol_test = vh.create('test')
        vh.remove(vol_test)
        with self.assertRaises(docker.errors.NotFound):
            vol_test.reload()

    def test_custom_namespace(self):
        """
        When the helper has a custom namespace, the volumes created are
        prefixed with the namespace.
        """
        vh = self.make_helper(namespace='integ')

        vol = vh.create('vol')
        self.addCleanup(vh.remove, vol)
        self.assertEqual(vol.name, 'integ_vol')


@dockertest()
class TestContainerHelper(unittest.TestCase):
    def setUp(self):
        self.client = docker.client.from_env()
        self.addCleanup(self.client.api.close)

        self.ih = ImageHelper(self.client)
        self.nh = NetworkHelper(self.client, 'test')
        self.addCleanup(self.nh._teardown)
        self.vh = VolumeHelper(self.client, 'test')
        self.addCleanup(self.vh._teardown)

    def make_helper(self, namespace='test'):
        ch = ContainerHelper(self.client, namespace, self.ih, self.nh, self.vh)
        self.addCleanup(ch._teardown)
        return ch

    def list_containers(self, *args, namespace='test', **kw):
        return filter_by_name(
            self.client.containers.list(*args, **kw), '{}_'.format(namespace))

    def test_teardown(self):
        """
        ContainerHelper._teardown() will remove any containers that were
        created, no matter what state they are in or even whether they still
        exist.
        """
        ch = self.make_helper()
        self.assertEqual([], self.list_containers(all=True))
        con_created = ch.create('created', IMG)
        self.assertEqual(con_created.status, 'created')

        con_running = ch.create('running', IMG)
        con_running.start()
        con_running.reload()
        self.assertEqual(con_running.status, 'running')

        con_stopped = ch.create('stopped', IMG)
        con_stopped.start()
        con_stopped.reload()
        self.assertEqual(con_stopped.status, 'running')
        con_stopped.stop()
        con_stopped.reload()
        self.assertNotEqual(con_stopped.status, 'running')

        con_removed = ch.create('removed', IMG)
        # We remove this behind the helper's back so the helper thinks it still
        # exists at teardown time.
        con_removed.remove()
        with self.assertRaises(docker.errors.NotFound):
            con_removed.reload()

        self.assertEqual(
            set([con_created, con_running, con_stopped]),
            set(self.list_containers(all=True)))

        with self.assertLogs('seaworthy', level='WARNING') as cm:
            ch._teardown()
        self.assertEqual(sorted(l.getMessage() for l in cm.records), [
            "Container 'test_created' still existed during teardown",
            "Container 'test_running' still existed during teardown",
            "Container 'test_stopped' still existed during teardown",
        ])
        self.assertEqual([], self.list_containers(all=True))

    def test_create(self):
        """
        We can create a container with various parameters without starting it.
        """
        ch = self.make_helper()

        con_simple = ch.create('simple', IMG)
        self.addCleanup(ch.remove, con_simple)
        self.assertEqual(con_simple.status, 'created')
        self.assertEqual(con_simple.attrs['Path'], 'nginx')

        con_cmd = ch.create('cmd', IMG, command='echo hello')
        self.addCleanup(ch.remove, con_cmd)
        self.assertEqual(con_cmd.status, 'created')
        self.assertEqual(con_cmd.attrs['Path'], 'echo')

        con_env = ch.create('env', IMG, environment={'FOO': 'bar'})
        self.addCleanup(ch.remove, con_env)
        self.assertEqual(con_env.status, 'created')
        self.assertIn('FOO=bar', con_env.attrs['Config']['Env'])

    def test_network(self):
        """
        When a container is created, the network settings are respected, and if
        no network settings are specified, the container is connected to a
        default network.
        """
        ch = self.make_helper()

        # When 'network' is provided, that network is used
        custom_network = self.nh.create('network')
        self.addCleanup(self.nh.remove, custom_network)
        con_network = ch.create('network', IMG, network=custom_network)
        self.addCleanup(ch.remove, con_network)
        networks = con_network.attrs['NetworkSettings']['Networks']
        self.assertEqual(list(networks.keys()), [custom_network.name])
        network = networks[custom_network.name]
        self.assertCountEqual(
            network['Aliases'], [con_network.id[:12], 'network'])

        # When 'network_mode' is provided, the default network is not used
        con_mode = ch.create('mode', IMG, network_mode='none')
        self.addCleanup(ch.remove, con_mode)
        networks = con_mode.attrs['NetworkSettings']['Networks']
        self.assertEqual(list(networks.keys()), ['none'])

        # When 'network_disabled' is True, the default network is not used
        con_disabled = ch.create(
            'disabled', IMG, network_disabled=True)
        self.addCleanup(ch.remove, con_disabled)
        self.assertEqual(con_disabled.attrs['NetworkSettings']['Networks'], {})

        con_default = ch.create('default', IMG)
        self.addCleanup(ch.remove, con_default)
        default_network_name = self.nh.get_default().name
        networks = con_default.attrs['NetworkSettings']['Networks']
        self.assertEqual(list(networks.keys()), [default_network_name])
        network = networks[default_network_name]
        self.assertCountEqual(
            network['Aliases'], [con_default.id[:12], 'default'])

    def test_network_by_id(self):
        """
        When a container is created, a network can be specified using the ID
        string for a network.
        """
        ch = self.make_helper()

        # When 'network' is provided as an ID, that network is used
        net_id = self.nh.create('id')
        self.addCleanup(self.nh.remove, net_id)
        con_id = ch.create('id', IMG, network=net_id.id)
        self.addCleanup(ch.remove, con_id)
        networks = con_id.attrs['NetworkSettings']['Networks']
        self.assertEqual(list(networks.keys()), [net_id.name])
        network = networks[net_id.name]
        self.assertCountEqual(network['Aliases'], [con_id.id[:12], 'id'])

    def test_network_by_short_id(self):
        """
        When a container is created, a network can be specified using the short
        ID string for a network.
        """
        ch = self.make_helper()

        # When 'network' is provided as a short ID, that network is used
        net_short_id = self.nh.create('short_id')
        self.addCleanup(self.nh.remove, net_short_id)
        con_short_id = ch.create(
            'short_id', IMG, network=net_short_id.short_id)
        self.addCleanup(ch.remove, con_short_id)
        networks = con_short_id.attrs['NetworkSettings']['Networks']
        self.assertEqual(list(networks.keys()), [net_short_id.name])
        network = networks[net_short_id.name]
        self.assertCountEqual(
            network['Aliases'], [con_short_id.id[:12], 'short_id'])

    def test_network_by_name(self):
        """
        When a container is created, a network can be specified using the name
        of a network.
        """
        ch = self.make_helper()

        # When 'network' is provided as a name, that network is used
        net_name = self.nh.create('name')
        self.addCleanup(self.nh.remove, net_name)
        con_name = ch.create('name', IMG, network=net_name.name)
        self.addCleanup(ch.remove, con_name)
        networks = con_name.attrs['NetworkSettings']['Networks']
        self.assertEqual(list(networks.keys()), [net_name.name])
        network = networks[net_name.name]
        self.assertCountEqual(network['Aliases'], [con_name.id[:12], 'name'])

    def test_network_by_invalid_type(self):
        """
        When a container is created, an error is raised if a network is
        specified using an invalid type.
        """
        ch = self.make_helper()

        with self.assertRaises(TypeError) as cm:
            ch.create('invalid_type', IMG, network=42)

        self.assertEqual(
            str(cm.exception),
            "Unexpected type <class 'int'>, expected <class 'str'> or <class "
            "'docker.models.networks.Network'>")

    def test_volumes(self):
        """
        When a container is created, a volume can be specified to be mounted.
        The container metadata should correctly identify the volumes and its
        mountpoint.
        """
        ch = self.make_helper()

        vol_test = self.vh.create('test')
        self.addCleanup(self.vh.remove, vol_test)
        con_volumes = ch.create(
            'volumes', IMG, volumes={vol_test: {'bind': '/vol', 'mode': 'rw'}})
        self.addCleanup(ch.remove, con_volumes)
        mounts = con_volumes.attrs['Mounts']
        self.assertEqual(len(mounts), 1)
        [mount] = mounts
        self.assertEqual(mount['Type'], 'volume')
        self.assertEqual(mount['Name'], vol_test.name)
        self.assertEqual(mount['Source'], vol_test.attrs['Mountpoint'])
        self.assertEqual(mount['Driver'], vol_test.attrs['Driver'])
        self.assertEqual(mount['Destination'], '/vol')
        self.assertEqual(mount['Mode'], 'rw')

    def test_volumes_short_form(self):
        """
        When a container is created, a volume can be specified to be mounted
        using a short-form bind specfier. The mode can be specified, but if not
        it defaults to read/write.
        """
        ch = self.make_helper()

        # Default mode: rw
        vol_default = self.vh.create('default')
        self.addCleanup(self.vh.remove, vol_default)
        con_default = ch.create('default', IMG, volumes={vol_default: '/vol'})
        self.addCleanup(ch.remove, con_default)
        mounts = con_default.attrs['Mounts']
        self.assertEqual(len(mounts), 1)
        [mount] = mounts
        self.assertEqual(mount['Type'], 'volume')
        self.assertEqual(mount['Name'], vol_default.name)
        self.assertEqual(mount['Destination'], '/vol')
        self.assertEqual(mount['Mode'], 'rw')

        # Specific mode: ro
        vol_mode = self.vh.create('mode')
        self.addCleanup(self.vh.remove, vol_mode)
        con_mode = ch.create('mode', IMG, volumes={vol_mode: '/mnt:ro'})
        self.addCleanup(ch.remove, con_mode)
        mounts = con_mode.attrs['Mounts']
        self.assertEqual(len(mounts), 1)
        [mount] = mounts
        self.assertEqual(mount['Type'], 'volume')
        self.assertEqual(mount['Name'], vol_mode.name)
        self.assertEqual(mount['Destination'], '/mnt')
        self.assertEqual(mount['Mode'], 'ro')

    def test_volumes_bind(self):
        """
        When a container is created, a bind mount can be specified in the
        ``volumes`` hash. The container metadata should correctly describe the
        bind mount.
        """
        ch = self.make_helper()

        tmpdir = tempfile.TemporaryDirectory()
        self.addCleanup(tmpdir.cleanup)
        con_bind = ch.create(
            'bind', IMG, volumes={tmpdir.name: {'bind': '/vol', 'mode': 'rw'}})
        self.addCleanup(ch.remove, con_bind)
        mounts = con_bind.attrs['Mounts']
        self.assertEqual(len(mounts), 1)
        [mount] = mounts
        self.assertEqual(mount['Type'], 'bind')
        self.assertEqual(mount['Source'], tmpdir.name)
        self.assertEqual(mount['Destination'], '/vol')
        self.assertTrue(mount['RW'])

    def test_volumes_by_name(self):
        """
        When a container is created, volumes can be specified using volume
        names. Volumes don't have IDs, only names :-/
        """
        ch = self.make_helper()

        # When 'volumes' is provided as a mapping from names, those volumes are
        # used
        vol_name = self.vh.create('name')
        self.addCleanup(self.vh.remove, vol_name)
        con_name = ch.create(
            'name', IMG,
            volumes={vol_name.name: {'bind': '/vol', 'mode': 'rw'}})
        self.addCleanup(ch.remove, con_name)
        mounts = con_name.attrs['Mounts']
        self.assertEqual(len(mounts), 1)
        [mount] = mounts
        self.assertEqual(mount['Name'], vol_name.name)

    def test_volumes_by_invalid_type(self):
        """
        When a container is created, an error is raised if a volume is
        specified using an invalid type.
        """
        ch = self.make_helper()

        with self.assertRaises(TypeError) as cm:
            ch.create('invalid_type', IMG,
                      volumes={42: {'bind': '/vol', 'mode': 'rw'}})

        self.assertEqual(
            str(cm.exception),
            "Unexpected type <class 'int'>, expected <class 'str'> or <class "
            "'docker.models.volumes.Volume'>")

    def test_volumes_specified_twice(self):
        """
        When a container is created, an error is raised if the same volume is
        specified twice: once using its string ID, and once using its model
        object.
        """
        ch = self.make_helper()

        # When 'volumes' is provided as a mapping from names, those volumes are
        # used
        vol_duplicate = self.vh.create('duplicate')
        self.addCleanup(self.vh.remove, vol_duplicate)
        with self.assertRaises(ValueError) as cm:
            ch.create(
                'duplicate', IMG,
                volumes={
                    vol_duplicate.name: {'bind': '/vol', 'mode': 'rw'},
                    vol_duplicate: {'bind': '/vol2', 'mode': 'ro'},
                })

        self.assertEqual(str(cm.exception),
                         "Volume 'test_duplicate' specified more than once")

    def test_remove(self):
        """
        We can remove a not-running container.
        """
        ch = self.make_helper()

        con_created = ch.create('created', IMG)
        self.assertEqual(con_created.status, 'created')
        ch.remove(con_created)
        with self.assertRaises(docker.errors.NotFound):
            con_created.reload()

        con_stopped = ch.create('stopped', IMG)
        con_stopped.start()
        con_stopped.stop()
        con_stopped.reload()
        self.assertEqual(con_stopped.status, 'exited')
        ch.remove(con_stopped)
        with self.assertRaises(docker.errors.NotFound):
            con_stopped.reload()

    def test_remove_force(self):
        """
        We can't remove a running container without forcing it.
        """
        ch = self.make_helper()

        con_running = ch.create('running', IMG)
        con_running.start()
        con_running.reload()
        self.assertEqual(con_running.status, 'running')
        with self.assertRaises(docker.errors.APIError):
            ch.remove(con_running, force=False)
        ch.remove(con_running)
        with self.assertRaises(docker.errors.NotFound):
            con_running.reload()

    def test_custom_namespace(self):
        """
        When the helper has a custom namespace, the containers created are
        prefixed with the namespace.
        """
        ch = self.make_helper(namespace='integ')

        con = ch.create('con', IMG, network_mode='none')
        self.addCleanup(ch.remove, con)
        self.assertEqual(con.name, 'integ_con')


@dockertest()
class TestDockerHelper(unittest.TestCase):
    def setUp(self):
        self.client = docker.client.from_env()
        self.addCleanup(self.client.api.close)

    def make_helper(self, *args, **kwargs):
        """
        Create and return a DockerHelper instance that will be cleaned up after
        the test.
        """
        dh = DockerHelper(*args, **kwargs)
        self.addCleanup(dh.teardown)
        return dh

    def test_custom_client(self):
        """
        When the DockerHelper is created with a custom client, that client is
        used.
        """
        client = docker.DockerClient(base_url='unix://var/run/docker.sock')
        dh = self.make_helper(client=client)

        self.assertIs(dh._client, client)

    def test_teardown_safe(self):
        """
        DockerHelper.teardown() is safe to call multiple times.

        There are no assertions here. We only care that calling teardown never
        raises any exceptions.
        """
        dh = self.make_helper()
        # These should silently do nothing.
        dh.teardown()
        dh.teardown()

    def test_remove_network_connected_to_created_container(self):
        """
        We can remove a network when it is connected to a container if the
        container hasn't been started yet.
        """
        dh = self.make_helper()

        net_test = dh.networks.create('test')

        con_created = dh.containers.create('created', IMG, network=net_test)
        self.addCleanup(dh.containers.remove, con_created)
        self.assertEqual(con_created.status, 'created')
        networks = con_created.attrs['NetworkSettings']['Networks']
        self.assertEqual(list(networks.keys()), [net_test.name])

        # Container not yet started so not listed as container connected to
        # network
        net_test.reload()
        self.assertEqual(net_test.containers, [])

        # Container not yet started so we can remove the network
        dh.networks.remove(net_test)
        with self.assertRaises(docker.errors.NotFound):
            net_test.reload()

        # The container will still think it's connected to the network
        con_created.reload()
        networks = con_created.attrs['NetworkSettings']['Networks']
        self.assertEqual(list(networks.keys()), [net_test.name])
        # But... we can't disconnect it from the old one

    def test_cannot_remove_network_connected_to_running_container(self):
        """
        We cannot remove a network when it is connected to a container if the
        container has been started. Once the container is disconnected, the
        network can be removed.
        """
        dh = self.make_helper()

        # Create a network and connect it to a container
        net_test = dh.networks.create('test')
        con_running = dh.containers.create('running', IMG, network=net_test)
        self.addCleanup(dh.containers.remove, con_running)
        networks = con_running.attrs['NetworkSettings']['Networks']
        self.assertEqual(list(networks.keys()), [net_test.name])

        # Start the container, now the network should know about it
        con_running.start()
        con_running.reload()
        self.assertEqual(con_running.status, 'running')
        net_test.reload()
        self.assertEqual(net_test.containers, [con_running])

        with self.assertRaises(docker.errors.APIError) as cm:
            dh.networks.remove(net_test)
        self.assertIn('network', str(cm.exception))
        self.assertIn('has active endpoints', str(cm.exception))

        # Once the container is disconnected, the network can be removed
        net_test.disconnect(con_running)
        dh.networks.remove(net_test)

        with self.assertRaises(docker.errors.NotFound):
            net_test.reload()

    def test_remove_network_connected_to_stopped_container(self):
        """
        We can remove a network when it is connected to a container if the
        container has been stopped.
        """
        dh = self.make_helper()

        # Create a network and connect it to a container
        net_test = dh.networks.create('test')
        con_stopped = dh.containers.create('stopped', IMG, network=net_test)
        self.addCleanup(dh.containers.remove, con_stopped)
        networks = con_stopped.attrs['NetworkSettings']['Networks']
        self.assertEqual(list(networks.keys()), [net_test.name])

        # Stop the container
        con_stopped.start()
        con_stopped.stop()
        con_stopped.reload()
        self.assertEqual(con_stopped.status, 'exited')

        net_test.reload()
        self.assertEqual(net_test.containers, [])
        dh.networks.remove(net_test)

        with self.assertRaises(docker.errors.NotFound):
            net_test.reload()

    def test_cannot_remove_mounted_volume(self):
        """
        We can't remove a volume mounted to a container no matter what state
        the container is in. Once the container has been removed, we can remove
        the volume.
        """
        dh = self.make_helper()

        vol_created = dh.volumes.create('created')
        con_created = dh.containers.create(
            'created', IMG,
            volumes={vol_created: {'bind': '/vol', 'mode': 'rw'}})
        self.assertEqual(con_created.status, 'created')

        # Try remove the volume... we can't
        with self.assertRaises(docker.errors.APIError) as cm:
            dh.volumes.remove(vol_created)
        self.assertIn('volume is in use', str(cm.exception))

        # Remove the container and then the volume
        dh.containers.remove(con_created)
        dh.volumes.remove(vol_created)

        vol_running = dh.volumes.create('running')
        con_running = dh.containers.create(
            'running', IMG,
            volumes={vol_running: {'bind': '/vol', 'mode': 'rw'}})

        con_running.start()
        con_running.reload()
        self.assertEqual(con_running.status, 'running')

        # Try remove the volume... we can't
        with self.assertRaises(docker.errors.APIError) as cm:
            dh.volumes.remove(vol_running)
        self.assertIn('volume is in use', str(cm.exception))

        # Remove the container and then the volume
        dh.containers.remove(con_running)
        dh.volumes.remove(vol_running)

        vol_stopped = dh.volumes.create('stopped')
        con_stopped = dh.containers.create(
            'stopped', IMG,
            volumes={vol_stopped: {'bind': '/vol', 'mode': 'rw'}})

        con_stopped.start()
        con_stopped.stop()
        con_stopped.reload()
        self.assertEqual(con_stopped.status, 'exited')

        # Try remove the volume... we can't
        with self.assertRaises(docker.errors.APIError) as cm:
            dh.volumes.remove(vol_stopped)
        self.assertIn('volume is in use', str(cm.exception))

        # Remove the container and then the volume
        dh.containers.remove(con_stopped)
        dh.volumes.remove(vol_stopped)

    def test_cannot_force_remove_mounted_volume(self):
        """
        We can't remove a volume mounted to a container even if we use
        ``force=True``.

        The Docker Engine API reference doesn't describe the force flag in much
        detail:
        "Force the removal of the volume"
        https://docs.docker.com/engine/api/v1.32/#operation/VolumeDelete

        The Docker Python client docs describe the force flag as:
        "Force removal of volumes that were already removed out of band by the
        volume driver plugin."
        https://docker-py.readthedocs.io/en/2.5.1/volumes.html#docker.models.volumes.Volume.remove
        """
        dh = self.make_helper()

        vol_test = dh.volumes.create('test')
        self.addCleanup(dh.volumes.remove, vol_test)
        con_created = dh.containers.create(
            'created', IMG, volumes={vol_test: {'bind': '/vol', 'mode': 'rw'}})
        self.addCleanup(dh.containers.remove, con_created)
        self.assertEqual(con_created.status, 'created')

        # Try remove the volume... we can't
        with self.assertRaises(docker.errors.APIError) as cm:
            dh.volumes.remove(vol_test, force=True)
        self.assertIn('volume is in use', str(cm.exception))

    def test_remove_container_with_volume(self):
        """
        When a volume is mounted to a container, and the container is removed,
        the volume itself is not removed.
        """
        dh = self.make_helper()

        vol_removed = dh.volumes.create('removed')
        self.addCleanup(dh.volumes.remove, vol_removed)
        con_removed = dh.containers.create(
            'removed', IMG,
            volumes={vol_removed: {'bind': '/vol', 'mode': 'rw'}})

        dh.containers.remove(con_removed)
        with self.assertRaises(docker.errors.NotFound):
            con_removed.reload()

        # The volume still exists: we can fetch it
        vol_removed.reload()

    def test_default_namespace(self):
        """
        When the Docker helper has the default namespace, all the resource
        helpers are created with that namespace.
        """
        dh = self.make_helper()
        self.assertEqual(dh.containers.namespace, 'test')
        self.assertEqual(dh.networks.namespace, 'test')
        self.assertEqual(dh.volumes.namespace, 'test')

    def test_custom_namespace(self):
        """
        When the Docker helper has a custom namespace, all the resource helpers
        are created with that namespace.
        """
        dh = self.make_helper(namespace='integ')
        self.assertEqual(dh.containers.namespace, 'integ')
        self.assertEqual(dh.networks.namespace, 'integ')
        self.assertEqual(dh.volumes.namespace, 'integ')

    def test_helper_for_model(self):
        """
        The _helper_for_model method returns the correct helper for the given
        Docker model type, or raises an exception if the model is of an unknown
        type.
        """
        dh = self.make_helper()
        self.assertIs(
            dh._helper_for_model(models.containers.Container), dh.containers)
        self.assertIs(dh._helper_for_model(models.images.Image), dh.images)
        self.assertIs(
            dh._helper_for_model(models.networks.Network), dh.networks)
        self.assertIs(dh._helper_for_model(models.volumes.Volume), dh.volumes)

        with self.assertRaises(ValueError) as cm:
            dh._helper_for_model(models.plugins.Plugin)
        self.assertEqual(
            str(cm.exception),
            "Unknown model type <class 'docker.models.plugins.Plugin'>")



import docker

import pytest

from seaworthy.checks import docker_client
from seaworthy.definitions import (
    ContainerDefinition, NetworkDefinition, VolumeDefinition)
from seaworthy.helpers import DockerHelper, fetch_images
from seaworthy.pytest.checks import dockertest
from seaworthy.pytest.fixtures import (
    clean_container_fixtures, docker_helper_fixture, image_fetch_fixture,
    resource_fixture)


# FIXME 2018-12-06: https://github.com/praekeltfoundation/seaworthy/issues/84
pytestmark = pytest.mark.skipif(int(pytest.__version__.split('.')[0]) >= 4,
                                reason='tests incompatible with pytest 4')


IMG = 'nginx:alpine'


def setup_module():
    with docker_client() as client:
        fetch_images(client, [IMG])


# We redefine the docker_helper fixture here to have function scope so our
# tests don't conflict.
docker_helper = docker_helper_fixture(scope='function')


@dockertest()
class TestDockerHelperFixture:
    def test_setup_teardown(self):
        """
        The fixture should yield a setup helper, and afterwards tear down the
        helper.
        """
        fixture_gen = docker_helper()
        helper = next(fixture_gen)
        assert isinstance(helper, DockerHelper)

        # Test we can create a container; if we can the helper must be set
        # up
        container = helper.containers.create('test', IMG)

        # Test things are torn down
        with pytest.raises(StopIteration):
            next(fixture_gen)

        with pytest.raises(docker.errors.NotFound):
            container.reload()


@dockertest()
class TestImageFetchFixtureFunc:
    def test_create_fixture(self, docker_helper):
        """
        We can create an fixture that fetches an image and returns an image
        model.
        """
        fixture = image_fetch_fixture(IMG, name='image', scope='module')
        image = fixture(docker_helper)

        assert isinstance(image, docker.models.images.Image)
        assert IMG in image.tags


# Dependent fixtures for use in TestResourceFixtureFunc::test_dependencies
volume = VolumeDefinition('test')
volume_fixture = resource_fixture(volume, 'volume_test')
network = NetworkDefinition('test')
network_fixture = resource_fixture(network, 'network_test')


@dockertest()
class TestResourceFixtureFunc:
    def test_setup_teardown(self, request, docker_helper):
        """
        The fixture should yield a started container, and afterwards stop and
        remove the container.
        """
        fixture = resource_fixture(
            ContainerDefinition(name='test', image=IMG), 'test')
        fixture_gen = fixture(request, docker_helper)
        container = next(fixture_gen)

        assert isinstance(container, ContainerDefinition)
        assert container.inner().status == 'running'

        # Test things are torn down
        with pytest.raises(StopIteration):
            next(fixture_gen)

        # Container has been stopped and removed
        assert not container.created

    def test_dependencies(self, request, docker_helper):
        """
        When the fixture depends on other fixtures, those fixtures should be
        setup when the fixture is used.
        """
        container_fixture = resource_fixture(
            ContainerDefinition(name='test', image=IMG), 'container_test',
            dependencies=('volume_test', 'network_test'))
        fixture_gen = container_fixture(request, docker_helper)
        container = next(fixture_gen)

        assert isinstance(container, ContainerDefinition)
        assert container.status() == 'running'
        assert volume.created
        assert network.created

        # Test things are torn down
        with pytest.raises(StopIteration):
            next(fixture_gen)

        # Container has been stopped and removed
        assert not container.created
        # Can't assert dependent fixture teardown--only happens once test
        # method returns


@dockertest()
class TestCleanContainerFixturesFunc:
    def test_setup_teardown(self, request, docker_helper):
        """
        The fixture should yield a started container, and afterwards stop and
        remove the container.
        """
        raw_fixture, fixture = clean_container_fixtures(
            ContainerDefinition(name='test', image=IMG), 'test')
        fixture_gen = raw_fixture(request, docker_helper)
        # TODO: Assert on cleaning fixture
        container = next(fixture_gen)

        assert isinstance(container, ContainerDefinition)
        assert container.inner().status == 'running'

        # Test things are torn down
        with pytest.raises(StopIteration):
            next(fixture_gen)

        # Container has been stopped and removed
        assert not container.created


@dockertest()
class PytestFixtureMixin:
    def make_definition(self, name):
        raise NotImplementedError()  # pragma: no cover

    def test_setup_teardown(self, request, docker_helper):
        """
        The ``pytest_fixture()`` method should return a fixture. That fixture
        should yield the definition with its resource created and when yielded
        again, the resource should be removed.
        """
        fixture = self.make_definition('test').pytest_fixture('test')
        fixture_gen = fixture(request, docker_helper)
        definition = next(fixture_gen)

        # Resource has been created
        assert definition.created

        # Test things are torn down
        with pytest.raises(StopIteration):
            next(fixture_gen)

        # Resource has been removed
        assert not definition.created


class TestContainerDefinition(PytestFixtureMixin):
    def make_definition(self, name):
        return ContainerDefinition(name, IMG)

    def test_clean_fixtures(self, request, docker_helper):
        """
        The fixture returned by the ``pytest_clean_fixture()`` method should
        yield a started container, and afterwards stop and remove the
        container.
        """
        raw_fixture, fixture = ContainerDefinition(
            name='test', image=IMG).pytest_clean_fixtures('test')
        fixture_gen = raw_fixture(request, docker_helper)
        # TODO: Assert on cleaning fixture
        container = next(fixture_gen)

        assert isinstance(container, ContainerDefinition)
        assert container.inner().status == 'running'

        # Test things are torn down
        with pytest.raises(StopIteration):
            next(fixture_gen)

        # Container has been stopped and removed
        assert not container.created


class TestNetworkDefinition(PytestFixtureMixin):
    def make_definition(self, name):
        return NetworkDefinition(name)


class TestVolumeDefinition(PytestFixtureMixin):
    def make_definition(self, name):
        return VolumeDefinition(name)



"""
Wrappers over Docker resource types to aid in setup/teardown of and interaction
with Docker resources.
"""

import functools

from docker import models

from seaworthy.helpers import DockerHelper
from seaworthy.stream.logs import stream_logs, wait_for_logs_matching
from seaworthy.stream.matchers import RegexMatcher, UnorderedMatcher


# This is a hack to control our generated documentation. The value of the
# attribute is ignored, only its presence or absence can be detected by the
# apigen machinery.
__apigen_inherited_members__ = None


def deep_merge(*dicts):
    """
    Recursively merge all input dicts into a single dict.
    """
    result = {}
    for d in dicts:
        if not isinstance(d, dict):
            raise Exception('Can only deep_merge dicts, got {}'.format(d))
        for k, v in d.items():
            # Whenever the value is a dict, we deep_merge it. This ensures that
            # (a) we only ever merge dicts with dicts and (b) we always get a
            # deep(ish) copy of the dicts and are thus safe from accidental
            # mutations to shared state.
            if isinstance(v, dict):
                v = deep_merge(result.get(k, {}), v)
            result[k] = v
    return result


class _DefinitionBase:
    __model_type__ = None

    def __init__(self, name, create_kwargs=None, helper=None):
        self.name = name

        self._create_args = ()
        self._create_kwargs = {} if create_kwargs is None else create_kwargs

        self._helper = None
        self.set_helper(helper)

        self._inner = None

    def create(self, **kwargs):
        """
        Create an instance of this resource definition.

        Only one instance may exist at any given time.
        """
        if self.created:
            raise RuntimeError(
                '{} already created.'.format(self.__model_type__.__name__))

        kwargs = self.merge_kwargs(self._create_kwargs, kwargs)

        self._inner = self.helper.create(
            self.name, *self._create_args, **kwargs)

    def remove(self, **kwargs):
        """
        Remove an instance of this resource definition.
        """
        self.helper.remove(self.inner(), **kwargs)
        self._inner = None

    def setup(self, helper=None, **create_kwargs):
        """
        Setup this resource so that is ready to be used in a test. If the
        resource has already been created, this call does nothing.

        For most resources, this just involves creating the resource in Docker.

        :param helper:
            The resource helper to use, if one was not provided when this
            resource definition was created.
        :param **create_kwargs: Keyword arguments passed to :meth:`.create`.

        :returns:
            This definition instance. Useful for creating and setting up a
            resource in a single step::

                volume = VolumeDefinition('volly').setup(helper=docker_helper)
        """
        if self.created:
            return

        self.set_helper(helper)
        self.create(**create_kwargs)
        return self

    def teardown(self):
        """
        Teardown this resource so that it no longer exists in Docker. If the
        resource has already been removed, this call does nothing.

        For most resources, this just involves removing the resource in Docker.
        """
        if not self.created:
            return

        self.remove()

    def __enter__(self):
        return self.setup()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.teardown()

    @property
    def helper(self):
        if self._helper is None:
            raise RuntimeError('No helper set.')
        return self._helper

    def set_helper(self, helper):
        """
        .. todo::

            Document this.
        """
        # We don't want to "unset" in this method.
        if helper is None:
            return

        # Get the right kind of helper if given a DockerHelper
        if isinstance(helper, DockerHelper):
            helper = helper._helper_for_model(self.__model_type__)

        # We already have this one.
        if helper is self._helper:
            return
        if self._helper is None:
            self._helper = helper
        else:
            raise RuntimeError('Cannot replace existing helper.')

    def as_fixture(self, name=None):
        """
        A decorator to inject this container into a function as a test fixture.
        """
        if name is None:
            name = self.name

        def deco(f):
            @functools.wraps(f)
            def wrapper(*args, **kw):
                with self:
                    kw[name] = self
                    return f(*args, **kw)
            return wrapper
        return deco

    def inner(self):
        """
        :returns: the underlying Docker model object
        """
        if not self.created:
            raise RuntimeError(
                '{} not created yet.'.format(self.__model_type__.__name__))
        return self._inner

    @property
    def created(self):
        return self._inner is not None

    def base_kwargs(self):
        """
        Override this method to provide dynamically generated base kwargs for
        the resource.
        """
        return {}

    def merge_kwargs(self, default_kwargs, kwargs):
        """
        Override this method to merge kwargs differently.
        """
        return deep_merge(self.base_kwargs(), default_kwargs, kwargs)


class ContainerDefinition(_DefinitionBase):
    """
    This is the base class for container definitions. Instances (and instances
    of subclasses) are intended to be used both as test fixtures and as
    convenient objects for operating on containers being tested.

    .. todo::

        Document this properly.

    A container object may be used as a context manager to ensure proper setup
    and teardown of the container around the code that uses it::

        with ContainerDefinition('my_container', IMAGE, helper=ch) as c:
            assert c.status() == 'running'

    (Note that this only works if the container has a helper set and does not
    have a container created.)
    """

    __model_type__ = models.containers.Container
    WAIT_TIMEOUT = 10.0

    def __init__(self, name, image, wait_patterns=None, wait_timeout=None,
                 create_kwargs=None, helper=None):
        """
        :param name:
            The name for the container. The actual name of the container is
            namespaced by ContainerHelper. This name will be used as a network
            alias for the container.
        :param image: image tag to use
        :param list wait_patterns:
            Regex patterns to use when checking that the container has started
            successfully.
        :param wait_timeout:
            Number of seconds to wait for the ``wait_patterns``. Defaults to
            ``self.WAIT_TIMEOUT``.
        :param dict create_kwargs:
            Other kwargs to use when creating the container.
        :param seaworthy.helper.ContainerHelper helper:
            A ContainerHelper instance used to create containers.
        """
        super().__init__(name, create_kwargs=create_kwargs, helper=helper)

        self._create_args = (image,)
        if wait_patterns:
            self.wait_matchers = [RegexMatcher(p) for p in wait_patterns]
        else:
            self.wait_matchers = None
        if wait_timeout is not None:
            self.wait_timeout = wait_timeout
        else:
            self.wait_timeout = self.WAIT_TIMEOUT

        self._http_clients = []

    def setup(self, helper=None, **run_kwargs):
        """
        Creates the container, starts it, and waits for it to completely start.

        :param helper:
            The resource helper to use, if one was not provided when this
            container definition was created.
        :param **run_kwargs: Keyword arguments passed to :meth:`.run`.

        :returns:
            This container definition instance. Useful for creating and setting
            up a container in a single step::

                con = ContainerDefinition('conny', 'nginx').setup(helper=dh)
        """
        if self.created:
            return

        self.set_helper(helper)
        self.run(**run_kwargs)
        self.wait_for_start()
        return self

    def teardown(self):
        """
        Stop and remove the container if it exists.
        """
        while self._http_clients:
            self._http_clients.pop().close()
        if self.created:
            self.halt()

    def status(self):
        """
        Get the container's current status from Docker.

        If the container does not exist (before creation and after removal),
        the status is ``None``.
        """
        if not self.created:
            return None
        self.inner().reload()
        return self.inner().status

    def start(self):
        """
        Start the container. The container must have been created.
        """
        self.inner().start()
        self.inner().reload()

    def stop(self, timeout=5):
        """
        Stop the container. The container must have been created.

        :param timeout:
            Timeout in seconds to wait for the container to stop before sending
            a ``SIGKILL``. Default: 5 (half the Docker default)
        """
        self.inner().stop(timeout=timeout)
        self.inner().reload()

    def run(self, fetch_image=True, **kwargs):
        """
        Create the container and start it. Similar to ``docker run``.

        :param fetch_image:
            Whether to try pull the image if it's not found. The behaviour here
            is similar to ``docker run`` and this parameter defaults to
            ``True``.
        :param **kwargs: Keyword arguments passed to :meth:`.create`.
        """
        self.create(fetch_image=fetch_image, **kwargs)
        self.start()

    def wait_for_start(self):
        """
        Wait for the container to start.

        By default this will wait for the log lines matching the patterns
        passed in the ``wait_patterns`` parameter of the constructor using an
        UnorderedMatcher. For more advanced checks for container startup, this
        method should be overridden.
        """
        if self.wait_matchers:
            matcher = UnorderedMatcher(*self.wait_matchers)
            self.wait_for_logs_matching(matcher, timeout=self.wait_timeout)

    def halt(self, stop_timeout=5):
        """
        Stop the container and remove it. The opposite of :meth:`run`.
        """
        self.stop(timeout=stop_timeout)
        self.remove()

    def clean(self):
        """
        This method should "clean" the container so that it is in the same
        state as it was when it was started. It is up to the implementer of
        this method to decide how the container should be cleaned. See
        :func:`~seaworthy.pytest.fixtures.clean_container_fixtures` for how
        this can be used with pytest fixtures.
        """
        raise NotImplementedError()

    @property
    def ports(self):
        """
        The ports (exposed and published) of the container.
        """
        return self.inner().attrs['NetworkSettings']['Ports']

    def _host_port(self, port_spec, index):
        if port_spec not in self.ports:
            raise ValueError("Port '{}' is not exposed".format(port_spec))

        mappings = self.ports[port_spec]
        if mappings is None:
            raise ValueError(
                "Port '{}' is not published to the host".format(port_spec))

        mapping = mappings[index]
        return mapping['HostIp'], mapping['HostPort']

    def get_host_port(self, container_port, proto='tcp', index=0):
        """
        :param container_port: The container port.
        :param proto: The protocol ('tcp' or 'udp').
        :param index: The index of the mapping entry to return.
        :returns: A tuple of the interface IP and port on the host.
        """
        port_spec = '{}/{}'.format(container_port, proto)
        return self._host_port(port_spec, index)

    def get_first_host_port(self):
        """
        Get the first mapping of the first (lowest) container port that has a
        mapping. Useful when a container publishes only one port.

        Note that unlike the Docker API, which sorts ports lexicographically
        (e.g. ``90/tcp`` > ``8000/tcp``), we sort ports numerically so that the
        lowest port is always chosen.
        """
        mapped_ports = {p: m for p, m in self.ports.items() if m is not None}
        if not mapped_ports:
            raise RuntimeError('Container has no published ports')

        def sort_key(port_string):
            port, proto = port_string.split('/', 1)
            return int(port), proto
        firt_port_spec = sorted(mapped_ports.keys(), key=sort_key)[0]

        return self._host_port(firt_port_spec, 0)

    def get_logs(self, stdout=True, stderr=True, timestamps=False, tail='all',
                 since=None):
        """
        Get container logs.

        This method does not support streaming, use :meth:`stream_logs` for
        that.
        """
        return self.inner().logs(
            stdout=stdout, stderr=stderr, timestamps=timestamps, tail=tail,
            since=since)

    def stream_logs(self, stdout=True, stderr=True, tail='all', timeout=10.0):
        """
        Stream container output.
        """
        return stream_logs(
            self.inner(), stdout=stdout, stderr=stderr, tail=tail,
            timeout=timeout)

    def wait_for_logs_matching(self, matcher, timeout=10, encoding='utf-8',
                               **logs_kwargs):
        """
        Wait for logs matching the given matcher.
        """
        wait_for_logs_matching(
            self.inner(), matcher, timeout=timeout, encoding=encoding,
            **logs_kwargs)

    def http_client(self, port=None):
        """
        Construct an HTTP client for this container.
        """
        # Local import to avoid potential circularity.
        from seaworthy.client import ContainerHttpClient
        client = ContainerHttpClient.for_container(self, container_port=port)
        self._http_clients.append(client)
        return client


class NetworkDefinition(_DefinitionBase):
    """
    This is the base class for network definitions.

    .. todo::

        Document this properly.
    """
    __model_type__ = models.networks.Network


class VolumeDefinition(_DefinitionBase):
    """
    This is the base class for volume definitions.

    The following is an example of how ``VolumeDefinition`` can be used to
    attach volumes to a container::

        from seaworthy.definitions import ContainerDefinition

        class DjangoContainer(ContainerDefinition):
            IMAGE = "seaworthy-demo:django"
            WAIT_PATTERNS = (r"Booting worker",)

            def __init__(self, name, socket_volume, static_volume, db_url):
                super().__init__(name, self.IMAGE, self.WAIT_PATTERNS)
                self.socket_volume = socket_volume
                self.static_volume = static_volume
                self.db_url = db_url

            def base_kwargs(self):
                return {
                    "volumes": {
                        self.socket_volume.inner(): "/var/run/gunicorn",
                        self.static_volume.inner(): "/app/static:ro",
                    },
                    "environment": {"DATABASE_URL": self.db_url}
                }

        # Create definition instances
        socket_volume = VolumeDefinition("socket")
        static_volume = VolumeDefinition("static")
        django_container = DjangoContainer(
            "django", socket_volume, static_volume,
            postgresql_container.database_url())

        # Create pytest fixtures
        socket_volume_fixture = socket_volume.pytest_fixture("socket_volume")
        static_volume_fixture = static_volume.pytest_fixture("static_volume")
        django_fixture = django_container.pytest_fixture(
            "django_container",
            dependencies=[
                "socket_volume", "static_volume", "postgresql_container"])

    This example is explained in the `introductory blog post`_ and
    `demo repository`_.

    .. todo::

        Document this properly.

    .. _`introductory blog post`:
        https://medium.com/mobileforgood/patterns-for-continuous-integration-with-docker-on-travis-ci-ba7e3a5ca2aa
    .. _`demo repository`:
        https://github.com/JayH5/seaworthy-demo
    """
    __model_type__ = models.volumes.Volume



"""
This package contains "core" tests that don't depend on anything that isn't
already a non-optional dependency of Seaworthy itself.

We have strong opinions about the testing tools we use, and we understand that
other people may have equally strong opinions that differ from ours. For this
reason, we have decided that none of Seaworthy's core functionality will depend
on pytest, testtools, or anything else that might get in the way of how people
might wants to write their tests.

To make sure none of these dependencies accidentally creep in, all tests in
this package are run in an environment without any extra (or optional)
dependencies installed.
"""



"""
Tools for asserting on processes running in containers using ``ps``.
"""

import attr

from .utils import output_lines


class PsException(Exception):
    """
    Exception indicating problems operating on process lists and trees.
    """


@attr.s
class PsRow:
    """
    Representation of a process list entry, containing the details of a single
    process.
    """

    pid = attr.ib(converter=int)
    ppid = attr.ib(converter=int)
    ruser = attr.ib()
    args = attr.ib()

    @classmethod
    def columns(cls):
        """
        List the columns required to construct a suitable ``ps`` command.
        """
        return [a.name for a in attr.fields(cls)]


def list_container_processes(container):
    """
    List the processes running inside a container.
    We use an exec rather than `container.top()` because we want to run 'ps'
    inside the container. This is because we want to get PIDs and usernames in
    the container's namespaces. `container.top()` uses 'ps' from outside the
    container in the host's namespaces. Note that this requires the container
    to have a 'ps' that responds to the arguments we give it-- we use BusyBox's
    (Alpine's) 'ps' as a baseline for available functionality.
    :param container: the container to query
    :return: a list of PsRow objects
    """
    cmd = ['ps', 'ax', '-o', ','.join(PsRow.columns())]
    ps_lines = output_lines(container.exec_run(cmd))

    header = ps_lines.pop(0)
    # We can't trust the header alignment because different ps implementations
    # use different alignments, some of which depend on the alignment of the
    # columns. Instead, we assume that all columns are whitespace-separated and
    # that only the last column may contain spaces.
    maxsplit = len(header.strip().split()) - 1
    ps_entries = [line.strip().split(None, maxsplit) for line in ps_lines]

    # Convert to PsRows
    ps_rows = [PsRow(*entry) for entry in ps_entries]

    # Filter out the row for ps itself
    cmd_string = ' '.join(cmd)
    ps_rows = [row for row in ps_rows if row.args != cmd_string]

    return ps_rows


@attr.s
class PsTree:
    """
    Node in a process tree, linking a :class:`PsRow` to its child processes.
    """
    row = attr.ib()
    children = attr.ib(default=attr.Factory(list))

    def count(self):
        """
        Return the number of processes in this subtree.
        """
        return 1 + sum(row.count() for row in self.children)


def _build_process_subtree(ps_rows, ps_tree, pids_seen):
    for row in ps_rows:
        if row.ppid == ps_tree.row.pid:
            if row.pid in pids_seen:
                raise PsException("Duplicate pid found: {}".format(row.pid))
            pids_seen.add(row.pid)
            tree = PsTree(row=row, children=[])
            ps_tree.children.append(tree)
            _build_process_subtree(ps_rows, tree, pids_seen)


def build_process_tree(ps_rows):
    """
    Build a tree structure from a list of PsRow objects.
    :param ps_rows: a list of PsRow objects
    :return: a PsTree object
    """
    ps_tree = None
    for row in ps_rows:
        if row.ppid == 0:
            if ps_tree is not None:
                raise PsException("Too many process tree roots (ppid=0) found")
            ps_tree = PsTree(row)
    if ps_tree is None:
        raise PsException("No process tree root (ppid=0) found")
    _build_process_subtree(ps_rows, ps_tree, set([ps_tree.row.pid]))
    if ps_tree.count() < len(ps_rows):
        raise PsException("Unreachable processes detected")
    assert ps_tree.count() == len(ps_rows)
    return ps_tree


__all__ = [
    'build_process_tree', 'list_container_processes',
    'PsException', 'PsRow', 'PsTree',
]



"""
pytest mark to skip tests that require Docker.
"""

import pytest

from seaworthy.checks import docker_available


def dockertest():
    """
    Skip tests that require Docker to be available.

    This is a function that returns a decorator so that we don't run arbitrary
    Docker client code on import. Unlike :func:`seaworthy.checks.dockertest`,
    this implementation doesn't require :class:`unittest.TestCase`. It does,
    however, require pytest.
    """
    return pytest.mark.skipif(
        not docker_available(), reason='Docker not available.')



"""
Tests for seaworthy.ps module.

Please note that these are "core" tests and thus may not depend on anything
that isn't already a non-optional dependency of Seaworthy itself.
"""

import unittest

from docker.models.containers import ExecResult

from seaworthy.ps import (
    PsException, PsRow, PsTree, build_process_tree, list_container_processes)


def mkrow(pid, ppid, ruser='root', args=None):
    if args is None:
        args = 'args for pid {}'.format(pid)
    return PsRow(pid, ppid, ruser, args)


class TestPsRow(unittest.TestCase):
    def test_columns(self):
        """
        The PsRow class knows what columns it requires from ps.
        """
        self.assertEqual(PsRow.columns(), ['pid', 'ppid', 'ruser', 'args'])

    def test_fields(self):
        """
        A PsRow can be created from field values of various types.
        """
        ps_row = PsRow('1', '0', 'root', 'tini -- true')
        self.assertEqual(
            (ps_row.pid, ps_row.ppid, ps_row.ruser, ps_row.args),
            (1, 0, 'root', 'tini -- true'))

        ps_row = PsRow(1, 0, 'root', 'tini -- true')
        self.assertEqual(
            (ps_row.pid, ps_row.ppid, ps_row.ruser, ps_row.args),
            (1, 0, 'root', 'tini -- true'))


class FakePsContainer:
    """
    A container object stub that emits canned process lists.
    """

    def __init__(self, rows):
        self.rows = rows

    def exec_run(self, cmd):
        # We only allow our own ps command to be run.
        assert cmd == ['ps', 'ax', '-o', 'pid,ppid,ruser,args']
        return ExecResult(0, b'\n'.join(self.rows))


class TestListContainerProcessesFunc(unittest.TestCase):
    def test_no_processes(self):
        """
        If we somehow get no processes other than the ps command, we return an
        empty list.
        """
        con = FakePsContainer([
            b'PID   PPID  RUSER    COMMAND',
            b'    6     0 root     ps ax -o pid,ppid,ruser,args',
        ])
        self.assertEqual(list_container_processes(con), [])

    def test_header_alignment(self):
        """
        Different ps implementations use different alignments for the column
        headers.
        """
        ps_rows = [PsRow(1, 0, 'root', 'sleep 2')]

        busybox_con = FakePsContainer([
            b'PID   PPID  RUSER    COMMAND',
            b'    1     0 root     sleep 2',
            b'    6     0 root     ps ax -o pid,ppid,ruser,args',
        ])
        self.assertEqual(ps_rows, list_container_processes(busybox_con))

        debian_con = FakePsContainer([
            b'  PID  PPID RUSER    COMMAND',
            b'    1     0 root     sleep 2',
            b'    6     0 root     ps ax -o pid,ppid,ruser,args',
        ])
        self.assertEqual(ps_rows, list_container_processes(debian_con))

    def test_header_alignment_long_pids(self):
        """
        For some header alignments, the pids may extend beyond the left edge of
        the column headers.
        """
        ps_rows = [
            PsRow(1, 0, 'root', 'tini -- foo'),
            PsRow(10000, 1, 'root', 'foo'),
            PsRow(10001, 10000, 'root', 'bar'),
        ]

        busybox_con = FakePsContainer([
            b'PID   PPID  RUSER    COMMAND',
            b'    1     0 root     tini -- foo',
            b'10000     1 root     foo',
            b'10001 10000 root     bar',
            b'    6     0 root     ps ax -o pid,ppid,ruser,args',
        ])
        self.assertEqual(ps_rows, list_container_processes(busybox_con))

        debian_con = FakePsContainer([
            b'  PID  PPID RUSER    COMMAND',
            b'    1     0 root     tini -- foo',
            b'10000     1 root     foo',
            b'10001 10000 root     bar',
            b'    6     0 root     ps ax -o pid,ppid,ruser,args',
        ])
        self.assertEqual(ps_rows, list_container_processes(debian_con))

    def test_many_processes(self):
        """
        If there are a lot of processes running in a container, we return all
        of them.
        """
        con = FakePsContainer([
            b'  PID  PPID RUSER    COMMAND',
            b'    1     0 root     tini -- django-entrypoint.sh args',
            b'    6     1 django   gunicorn master args',
            b'   17     6 root     nginx master args',
            b'   18     6 django   celery worker args',
            b'   19     6 django   celery beat args',
            b'   26    17 nginx    nginx worker args',
            b'   34     6 django   gunicorn worker args',
            b'   48     0 root     ps ax -o pid,ppid,ruser,args',
        ])
        self.assertEqual(list_container_processes(con), [
            PsRow(1, 0, 'root', 'tini -- django-entrypoint.sh args'),
            PsRow(6, 1, 'django', 'gunicorn master args'),
            PsRow(17, 6, 'root', 'nginx master args'),
            PsRow(18, 6, 'django', 'celery worker args'),
            PsRow(19, 6, 'django', 'celery beat args'),
            PsRow(26, 17, 'nginx', 'nginx worker args'),
            PsRow(34, 6, 'django', 'gunicorn worker args'),
        ])


class TestPsTree(unittest.TestCase):
    def test_count(self):
        """
        A PsTree knows how many entries it contains.
        """
        self.assertEqual(1, PsTree(mkrow(1, 0)).count())

        self.assertEqual(3, PsTree(mkrow(1, 0), [
            PsTree(mkrow(6, 1), [
                PsTree(mkrow(8, 6)),
            ]),
        ]).count())

        self.assertEqual(6, PsTree(mkrow(1, 0), [
            PsTree(mkrow(6, 1), [
                PsTree(mkrow(8, 6)),
            ]),
            PsTree(mkrow(9, 1), [
                PsTree(mkrow(11, 9)),
                PsTree(mkrow(12, 9)),
            ]),
        ]).count())


class TestBuildProcessTreeFunc(unittest.TestCase):
    def test_single_process(self):
        """
        We can build a PsTree for a single process.
        """
        ps_row = PsRow('1', '0', 'root', 'tini -- echo "hi"')
        ps_tree = build_process_tree([ps_row])

        self.assertEqual(ps_tree, PsTree(ps_row, children=[]))

    def test_simple_tree(self):
        """
        We can build a PsTree for a list of grandparent/parent/child processes.
        """
        ps_rows = [
            mkrow(1, 0, 'root', "tini -- nginx -g 'daemon off;'"),
            mkrow(6, 1, 'root', 'nginx: master process nginx -g daemon off;'),
            mkrow(8, 6, 'nginx', 'nginx: worker process'),
        ]
        ps_tree = build_process_tree(ps_rows)
        self.assertEqual(ps_tree, PsTree(ps_rows[0], [
            PsTree(ps_rows[1], [
                PsTree(ps_rows[2], []),
            ]),
        ]))

    def test_bigger_tree(self):
        """
        We can build a PsTree for a more complicated process list.
        """
        ps_rows = [
            None,  # Dummy entry so list indices match pids.
            mkrow(1, 0),
            mkrow(2, 1),
            mkrow(3, 1),
            mkrow(4, 2),
            mkrow(5, 3),
            mkrow(6, 3),
            mkrow(7, 4),
            mkrow(8, 2),
            mkrow(9, 1),
        ]
        ps_tree = build_process_tree(ps_rows[1:])
        self.assertEqual(ps_tree, PsTree(ps_rows[1], [
            PsTree(ps_rows[2], [
                PsTree(ps_rows[4], [
                    PsTree(ps_rows[7]),
                ]),
                PsTree(ps_rows[8]),
            ]),
            PsTree(ps_rows[3], [
                PsTree(ps_rows[5]),
                PsTree(ps_rows[6]),
            ]),
            PsTree(ps_rows[9]),
        ]))

    def test_no_root_pid(self):
        """
        We can't build a process tree if we don't have a root process.
        """
        with self.assertRaises(PsException) as cm:
            build_process_tree([])
        self.assertIn("No process tree root", str(cm.exception))

        with self.assertRaises(PsException) as cm:
            build_process_tree([
                mkrow(2, 1),
                mkrow(3, 1),
                mkrow(4, 2),
            ])
        self.assertIn("No process tree root", str(cm.exception))

    def test_multiple_root_pids(self):
        """
        We can't build a process tree if we have too many root processes.
        """
        with self.assertRaises(PsException) as cm:
            build_process_tree([
                mkrow(1, 0),
                mkrow(2, 0),
                mkrow(4, 2),
            ])
        self.assertIn("Too many process tree roots", str(cm.exception))

    def test_malformed_process_tree(self):
        """
        We can't build a process tree with disconnected processes.
        """
        with self.assertRaises(PsException) as cm:
            build_process_tree([
                mkrow(1, 0),
                mkrow(2, 1),
                mkrow(4, 3),
            ])
        self.assertIn("Unreachable processes", str(cm.exception))

    def test_duplicate_pids(self):
        """
        We can't build a process tree with duplicate pids.
        """
        with self.assertRaises(PsException) as cm:
            build_process_tree([
                mkrow(1, 0),
                mkrow(2, 1),
                mkrow(2, 1),
                mkrow(3, 2),
            ])
        self.assertIn("Duplicate pid found: 2", str(cm.exception))



import time
import unittest
from datetime import datetime

from seaworthy.checks import docker_client, dockertest
from seaworthy.definitions import (
    ContainerDefinition, NetworkDefinition, VolumeDefinition)
from seaworthy.helpers import DockerHelper, fetch_images
from seaworthy.stream.matchers import EqualsMatcher

IMG_SCRIPT = 'alpine:latest'
IMG_WAIT = 'nginx:alpine'


@dockertest()
class DefinitionTestMixin:
    def _setup(self):
        self.dh = DockerHelper()
        self.addCleanup(self.dh.teardown)

        self.definition = self.with_cleanup(
            self.make_definition('test', helper=self.dh))
        self.helper = self.definition.helper

    def with_cleanup(self, definition):
        self.addCleanup(definition.teardown)
        return definition

    def make_definition(self, name, helper=None):
        raise NotImplementedError()  # pragma: no cover

    def test_helper_not_set(self):
        """
        By default, we have no helper.
        """
        no_helper = self.make_definition('no_helper')
        self.assertIsNone(no_helper._helper)
        with self.assertRaises(RuntimeError) as cm:
            no_helper.helper
        self.assertEqual(str(cm.exception), 'No helper set.')

    def test_helper_set_in_constructor(self):
        """
        We can set a helper in the constructor.
        """
        with_helper = self.make_definition('with_helper', helper=self.helper)
        self.assertIs(with_helper._helper, self.helper)
        self.assertIs(with_helper.helper, self.helper)

    def test_helper_set_to_none(self):
        """
        Setting helper to None has no effect even if we already have
        one.
        """
        no_helper = self.make_definition('no_helper')
        self.assertIsNone(no_helper._helper)
        no_helper.set_helper(None)
        self.assertIsNone(no_helper._helper)

        with_helper = self.make_definition('with_helper', helper=self.helper)
        self.assertIs(with_helper._helper, self.helper)
        with_helper.set_helper(None)
        self.assertIs(with_helper._helper, self.helper)

    def test_helper_set_to_current(self):
        """
        Setting helper to the one we already have has no effect.
        """
        with_helper = self.make_definition('with_helper', helper=self.helper)
        self.assertIs(with_helper._helper, self.helper)
        with_helper.set_helper(self.helper)
        self.assertIs(with_helper._helper, self.helper)

    def test_cannot_replace_helper(self):
        """
        If we already have a helper, we can't set a different one.
        """
        with_helper = self.make_definition('with_helper', helper=self.helper)
        self.assertIs(with_helper.helper, self.helper)
        # TODO: fix this...
        with self.assertRaises(RuntimeError) as cm:
            with_helper.set_helper(DockerHelper())
        self.assertEqual(
            str(cm.exception), 'Cannot replace existing helper.')
        self.assertIs(with_helper.helper, self.helper)

    def test_helper_set_to_docker_helper(self):
        """
        Setting helper to a DockerHelper instance gets us the correct helper.
        """
        with_helper = self.make_definition('with_helper', helper=self.dh)
        self.assertIs(
            with_helper._helper.collection.model, with_helper.__model_type__)
        self.assertIs(with_helper._helper, self.helper)

    def test_create_only_if_not_created(self):
        """
        The resource cannot be created more than once.
        """
        self.definition.create()

        # We can't create the resource when it's already created
        with self.assertRaises(RuntimeError) as cm:
            self.definition.create()
        self.assertRegex(str(cm.exception), r'^\w+ already created\.$')

        self.definition.remove()

    def test_remove_only_if_created(self):
        """
        The resource can only be removed if it has been created.
        """
        self.definition.create()

        # We can remove the resource if it's created
        self.definition.remove()

        with self.assertRaises(RuntimeError) as cm:
            self.definition.remove()
        self.assertRegex(str(cm.exception), r'^\w+ not created yet\.$')

    def test_inner_only_if_created(self):
        """
        We can only access the inner object if the resource has been
        created.
        """
        # If we try get the resource before it's created it'll fail
        with self.assertRaises(RuntimeError) as cm:
            self.definition.inner()
        self.assertRegex(str(cm.exception), r'^\w+ not created yet\.$')

        self.definition.create()

        # We can get the resource once it's created
        inner = self.definition.inner()
        self.assertIsInstance(inner, self.definition.__model_type__)

        self.definition.remove()
        with self.assertRaises(RuntimeError) as cm:
            self.definition.inner()
        self.assertRegex(str(cm.exception), r'^\w+ not created yet\.$')

    def test_setup_teardown(self):
        """
        We can use the ``setup`` and ``teardown`` methods to create and remove
        the resource.
        """
        self.assertFalse(self.definition.created)

        self.definition.setup()
        self.assertTrue(self.definition.created)

        self.definition.teardown()
        self.assertFalse(self.definition.created)

    def test_setup_kwargs(self):
        """
        We can pass keyword args to ``setup``.
        """
        self.assertFalse(self.definition.created)

        self.definition.setup(labels={'SETUP_KWARGS': 'working'})
        self.addCleanup(self.definition.teardown)
        self.assertTrue(self.definition.created)
        labels = self.definition.inner().attrs['Labels']
        self.assertEqual(labels, {'SETUP_KWARGS': 'working'})

    def test_setup_multiple_calls(self):
        """
        setup() can be called multiple times after it has been called once
        without having any effect.
        """
        self.assertFalse(self.definition.created)

        self.definition.setup()
        self.assertTrue(self.definition.created)
        inner = self.definition.inner()

        self.definition.setup()
        self.definition.setup()

        self.assertIs(inner, self.definition.inner())

    def test_teardown_multiple_calls(self):
        """
        teardown() can be called multiple times after it has been called once
        without having any effect.
        """
        self.assertFalse(self.definition.created)
        self.definition.setup()
        self.assertTrue(self.definition.created)

        self.definition.teardown()
        self.assertFalse(self.definition.created)

        self.definition.teardown()
        self.definition.teardown()

    def test_context_manager(self):
        """
        We can use a definition object as a context manager (which returns
        itself) to create and remove it.
        """
        self.assertFalse(self.definition.created)
        with self.definition as definition:
            self.assertIs(definition, self.definition)
            self.assertTrue(definition.created)
        self.assertFalse(self.definition.created)

    def test_fixture_on_function(self):
        """
        We can make a function decorator fixture for our definition.
        """
        self.assertFalse(self.definition.created)
        closure_state = []

        @self.definition.as_fixture()
        def foo(test):
            self.assertIs(test, self.definition)
            self.assertTrue(self.definition.created)
            closure_state.append(1)

        self.assertEqual(closure_state, [])
        self.assertFalse(self.definition.created)
        foo()
        self.assertFalse(self.definition.created)
        self.assertEqual(closure_state, [1])

    def test_fixture_on_method(self):
        """
        The fixture decorator works on methods too.
        """
        self.assertFalse(self.definition.created)
        closure_state = []

        class Foo:
            def __init__(self, value, tc):
                self.value = value
                self.tc = tc

            @self.definition.as_fixture()
            def foo(self, test):
                self.tc.assertIs(test, self.tc.definition)
                self.tc.assertTrue(test.created)
                closure_state.append(self.value)

        self.assertEqual(closure_state, [])
        self.assertFalse(self.definition.created)
        Foo(3, self).foo()
        self.assertFalse(self.definition.created)
        self.assertEqual(closure_state, [3])

    def test_fixture_with_name(self):
        """
        The fixture decorator can override the parameter name used.
        """
        self.assertFalse(self.definition.created)
        closure_state = []

        @self.definition.as_fixture(name='something_more_sensible')
        def foo(something_more_sensible):
            self.assertIs(something_more_sensible, self.definition)
            self.assertTrue(something_more_sensible.created)
            closure_state.append(1)

        self.assertEqual(closure_state, [])
        self.assertFalse(self.definition.created)
        foo()
        self.assertFalse(self.definition.created)
        self.assertEqual(closure_state, [1])

    def test_merge_kwargs(self):
        """
        The default merge_kwargs() method deep-merges the two kwargs dicts
        passed to it.
        """
        create_kwargs = {'a': {'aa': 1, 'ab': 2}, 's': 'foo', 't': 'bar'}
        kwargs = {'a': {'ba': 3, 'ab': 4}, 'r': 'arr', 't': 'baz'}
        self.assertEqual(self.definition.merge_kwargs(create_kwargs, kwargs), {
            'a': {'aa': 1, 'ab': 4, 'ba': 3},
            'r': 'arr',
            's': 'foo',
            't': 'baz',
        })

    def test_merge_kwargs_with_base(self):
        """
        The default merge_kwargs() method deep-merges the two kwargs dicts
        passed to it on top of the output of base_kwargs().
        """
        self.definition.base_kwargs = (
            lambda: {'a': {'aa': 0, 'bb': 6}, 'b': 'base'})
        create_kwargs = {'a': {'aa': 1, 'ab': 2}, 's': 'foo', 't': 'bar'}
        kwargs = {'a': {'ba': 3, 'ab': 4}, 'r': 'arr', 't': 'baz'}
        self.assertEqual(self.definition.merge_kwargs(create_kwargs, kwargs), {
            'a': {'aa': 1, 'ab': 4, 'ba': 3, 'bb': 6},
            'b': 'base',
            'r': 'arr',
            's': 'foo',
            't': 'baz',
        })

    def test_merge_kwargs_dicts_only(self):
        """
        The kwargs we merge must be dicts.
        """
        with self.assertRaises(Exception):
            self.definition.merge_kwargs({}, 'hello')
        with self.assertRaises(Exception):
            self.definition.merge_kwargs('hello', {})


@dockertest()
class TestContainerDefinition(unittest.TestCase, DefinitionTestMixin):
    @classmethod
    def setUpClass(cls):
        with docker_client() as client:
            fetch_images(client, [IMG_SCRIPT, IMG_WAIT])

    def setUp(self):
        self._setup()

    def make_definition(self, name, helper=None):
        return ContainerDefinition(name, IMG_WAIT, helper=helper)

    def test_setup_teardown(self):
        """
        We can use the ``setup`` and ``teardown`` methods to create and remove
        the container.
        """
        self.assertFalse(self.definition.created)

        self.definition.setup()
        self.assertTrue(self.definition.created)
        # Also assert that the container is running
        self.assertEqual(self.definition.status(), 'running')

        self.definition.teardown()
        self.assertFalse(self.definition.created)
        # No status for container now
        self.assertIs(self.definition.status(), None)

    def test_setup_kwargs(self):
        """
        We can pass keyword args to ``setup``.
        """
        self.assertFalse(self.definition.created)

        self.definition.setup(environment={'SETUP_KWARGS': 'working'})
        self.addCleanup(self.definition.teardown)
        self.assertTrue(self.definition.created)
        env = self.definition.inner().attrs['Config']['Env']
        self.assertIn('SETUP_KWARGS=working', env)

    def test_context_manager(self):
        """
        We can use a definition object as a context manager (which returns
        itself) to create and remove it.
        """
        self.assertFalse(self.definition.created)
        with self.definition as definition:
            self.assertIs(definition, self.definition)
            self.assertTrue(definition.created)
            # Also assert that the container is running
            self.assertEqual(definition.status(), 'running')
        self.assertFalse(self.definition.created)
        # No status for container now
        self.assertIs(self.definition.status(), None)

    def test_start(self):
        """
        We can start a container after creating it.
        """
        self.definition.create()
        inner = self.definition.inner()
        self.assertEqual(inner.status, 'created')

        self.definition.start()
        self.assertEqual(inner.status, 'running')

    def test_stop(self):
        """
        We can stop a running container.
        """
        # We don't test the timeout because that's just passed directly through
        # to Docker and it's nontrivial to construct a container that takes a
        # specific amount of time to stop.
        self.definition.create()
        inner = self.definition.inner()
        self.assertEqual(inner.status, 'created')

        self.definition.start()
        self.assertEqual(inner.status, 'running')

        self.definition.stop()
        self.assertEqual(inner.status, 'exited')

    def test_wait_timeout_default(self):
        """
        When wait_timeout isn't passed to the constructor, the default timeout
        is used.
        """
        container = ContainerDefinition('timeout', IMG_WAIT)
        self.assertEqual(
            container.wait_timeout, ContainerDefinition.WAIT_TIMEOUT)

    def test_wait_timeout_override(self):
        """
        When wait_timeout is passed to the constructor, it is used in place of
        the default.
        """
        timeout = ContainerDefinition.WAIT_TIMEOUT + 10.0
        container = ContainerDefinition(
            'timeout', IMG_WAIT, wait_timeout=timeout)
        self.assertEqual(container.wait_timeout, timeout)

    def test_default_clean(self):
        """By default, the ``clean`` method raises a NotImplementedError."""
        with self.assertRaises(NotImplementedError):
            self.definition.clean()

    def test_create_kwargs_handling(self):
        """
        The keyword args passed used for container creation come from the
        return value of merge_kwargs() called on create_kwargs from the
        constructor and kwargs from the create() method.
        """
        create_kwargs = {
            'environment': {'CREATE_KWARGS': 't', 'KWARGS_MERGED': 'f'},
        }
        kwargs = {'environment': {'KWARGS': 't', 'KWARGS_MERGED': 't'}}
        merge_kwargs_args = []

        class SubContainer(ContainerDefinition):
            def merge_kwargs(self, *args):
                merge_kwargs_args.extend(args)
                return super().merge_kwargs(*args)

        c = self.with_cleanup(SubContainer(
            'kwargs', IMG_WAIT, create_kwargs=create_kwargs,
            helper=self.helper))
        c.create(**kwargs)

        self.assertEqual(merge_kwargs_args, [create_kwargs, kwargs])
        c_env = [v for v in c.inner().attrs['Config']['Env'] if 'KWARGS' in v]
        self.assertEqual(
            sorted(c_env), ['CREATE_KWARGS=t', 'KWARGS=t', 'KWARGS_MERGED=t'])

    def test_ports(self):
        """
        We can get the ports exposed or published on a container.
        """
        self.definition.run(
            fetch_image=False, ports={'8000/tcp': ('127.0.0.1', '10701')})

        # We're not interested in the order of the ports
        self.assertCountEqual(self.definition.ports.items(), [
            ('80/tcp', None),
            ('8000/tcp', [{'HostIp': '127.0.0.1', 'HostPort': '10701'}]),
        ])

    def test_get_host_port(self):
        """
        We can get the host port mapping of a container.
        """
        self.definition.run(fetch_image=False, ports={
            '8080/tcp': ('127.0.0.1',),
            '9090/tcp': ('127.0.0.1', '10701'),
            '9191/udp': '10702',
        })

        # We get a random high port number here.
        host_iface, random_host_port = self.definition.get_host_port('8080')
        self.assertEqual(host_iface, '127.0.0.1')
        self.assertGreater(int(random_host_port), 1024)
        self.assertLess(int(random_host_port), 65536)

        # We get the specific port we defined here.
        host_iface, specific_host_port = self.definition.get_host_port('9090')
        self.assertEqual(host_iface, '127.0.0.1')
        self.assertEqual(specific_host_port, '10701')

        # We get a UDP port we defined.
        _, udp_host_port = self.definition.get_host_port('9191', proto='udp')
        self.assertEqual(udp_host_port, '10702')

        # FIXME: Don't bother testing index != 0, the port order is
        # unpredictable :-(

    def test_get_host_port_not_exposed(self):
        """
        When we try to get the host port for a container port that hasn't been
        exposed, an error is raised.
        """
        self.definition.run(
            fetch_image=False, ports={'8000/tcp': ('127.0.0.1', '10701')})

        self.assertNotIn('90/tcp', self.definition.ports)

        with self.assertRaises(ValueError) as cm:
            self.definition.get_host_port('90')
        self.assertEqual(str(cm.exception), "Port '90/tcp' is not exposed")

    def test_get_host_port_not_published(self):
        """
        When we try to get the host port for a container port that hasn't been
        published to the host, an error is raised.
        """
        self.definition.run(
            fetch_image=False, ports={'8000/tcp': ('127.0.0.1', '10701')})

        # The Nginx image EXPOSEs port 80, but we don't publish it
        self.assertIn('80/tcp', self.definition.ports)

        with self.assertRaises(ValueError) as cm:
            self.definition.get_host_port('80')
        self.assertEqual(
            str(cm.exception), "Port '80/tcp' is not published to the host")

    def test_get_first_host_port(self):
        """
        When we get the first host port for the container, the host port mapped
        to the lowest container port is returned.
        """
        self.definition.run(
            fetch_image=False, ports={
                '8000/tcp': ('127.0.0.1',),
                '90/tcp': ('127.0.0.1', '10701'),
                '90/udp': ('127.0.0.1', '10702'),
            })

        # The Nginx image EXPOSEs port 80, but it's not published so shouldn't
        # be considered by ``get_first_host_port()``
        self.assertIn('80/tcp', self.definition.ports)

        host_iface, host_port = self.definition.get_first_host_port()
        self.assertEqual(host_iface, '127.0.0.1')
        self.assertEqual(host_port, '10701')

    def test_get_first_host_port_no_mappings(self):
        """
        When we try to get the first host port, but the container has no
        published ports, an error is raised.
        """
        self.definition.run(fetch_image=False)

        # The Nginx image EXPOSEs port 80, but it's not published so shouldn't
        # be considered by ``get_first_host_port()``
        self.assertIn('80/tcp', self.definition.ports)

        with self.assertRaises(RuntimeError) as cm:
            self.definition.get_first_host_port()

        self.assertEqual(str(cm.exception), 'Container has no published ports')

    def run_logs_container(self, logs, wait=True, delay=0.01):
        # Sleep some amount between lines to ensure ordering across stdout and
        # stderr.
        script = '\nsleep {}\n'.format(delay).join(logs)

        script_con = self.with_cleanup(
            ContainerDefinition('script', IMG_SCRIPT, helper=self.helper))

        script_con.run(fetch_image=False, command=['sh', '-c', script])
        # Wait for the output to arrive.
        if wait:
            # Wait a minimum of 100ms to avoid jitter with small intervals.
            time.sleep(max(0.1, len(logs) * delay))
        return script_con

    def test_get_logs_out_err(self):
        """
        We can choose stdout and/or stderr when getting logs from a container.
        """
        script = self.run_logs_container([
            'echo "o0"', 'echo "e0" >&2',
            'echo "o1"', 'echo "e1" >&2',
        ])

        self.assertEqual(script.get_logs(), b'o0\ne0\no1\ne1\n')
        self.assertEqual(script.get_logs(stdout=False), b'e0\ne1\n')
        self.assertEqual(script.get_logs(stderr=False), b'o0\no1\n')

    def test_get_logs_tail(self):
        """
        We can choose how many lines to tail when getting logs from a
        container.

        NOTE: Lines are tailed *before* stdout/stderr are filtered out.
        """
        script = self.run_logs_container([
            'echo "o0"', 'echo "e0" >&2',
            'echo "o1"', 'echo "e1" >&2',
        ])

        self.assertEqual(script.get_logs(tail='all'), b'o0\ne0\no1\ne1\n')
        self.assertEqual(script.get_logs(tail=10000), b'o0\ne0\no1\ne1\n')
        self.assertEqual(script.get_logs(tail=0), b'')
        self.assertEqual(script.get_logs(tail=1), b'e1\n')
        self.assertEqual(script.get_logs(tail=2), b'o1\ne1\n')
        # The entries are tailed *before* they're filtered. :-(
        self.assertEqual(script.get_logs(stderr=False, tail=1), b'')
        self.assertEqual(script.get_logs(stderr=False, tail=2), b'o1\n')
        self.assertEqual(script.get_logs(stderr=False, tail=3), b'o1\n')
        self.assertEqual(script.get_logs(stderr=False, tail=4), b'o0\no1\n')

    def test_get_logs_timestamps(self):
        """
        We can ask for timestamps on our logs.
        """
        before = datetime.utcnow()
        script = self.run_logs_container([
            'echo "o0"', 'echo "e0" >&2',
            'echo "o1"', 'echo "e1" >&2',
        ])
        after = datetime.utcnow()

        self.assertEqual(script.get_logs(), b'o0\ne0\no1\ne1\n')
        raw_lines = script.get_logs(timestamps=True).splitlines()

        earlier = before
        for line, expected in zip(raw_lines, [b'o0', b'e0', b'o1', b'e1']):
            ts, ln = line.split(b' ', 1)
            self.assertEqual(ln, expected)
            # Truncate the nanoseconds, because we can't parse them.
            ts = (ts[:26] + ts[-1:]).decode('utf8')
            ts = datetime.strptime(ts, '%Y-%m-%dT%H:%M:%S.%fZ')
            self.assertLess(earlier, ts)
            earlier = ts
        self.assertLess(earlier, after)

    def test_stream_logs_timeout(self):
        """
        We can stream logs with a timeout.
        """
        script = self.run_logs_container([
            'echo "o0"', 'echo "e0" >&2',
            'echo "o1"', 'echo "e1" >&2',
        ], delay=0.25, wait=False)

        time.sleep(0.01)
        lines = []
        with self.assertRaises(TimeoutError):
            for line in script.stream_logs(tail=0, timeout=0.3):
                lines.append(line)
        # We missed the first line and time out before the third.
        self.assertEqual(lines, [b'e0\n'])

        lines = []
        for line in script.stream_logs(tail=0, timeout=1.0):
            lines.append(line)
        # We should get the third and fourth now.
        self.assertEqual(lines, [b'o1\n', b'e1\n'])

    def test_stream_tail_all(self):
        """
        We can stream logs, including old logs, with a timeout.
        """
        script = self.run_logs_container([
            'echo "o0"', 'echo "e0" >&2',
            'echo "o1"', 'echo "e1" >&2',
        ], delay=0.25, wait=False)

        time.sleep(0.01)
        lines = []
        with self.assertRaises(TimeoutError):
            for line in script.stream_logs(tail='all', timeout=0.3):
                lines.append(line)
        # We get the (old) first line and time out before the third.
        self.assertEqual(lines, [b'o0\n', b'e0\n'])

        lines = []
        for line in script.stream_logs(tail='all', timeout=1.0):
            lines.append(line)
        self.assertEqual(lines, [b'o0\n', b'e0\n', b'o1\n', b'e1\n'])

    def test_stream_logs_stdout(self):
        """
        We can choose stdout when streaming logs.
        """
        script = self.run_logs_container([
            'echo "first"',
            'echo "o0"', 'echo "e0" >&2',
            'echo "o1"', 'echo "e1" >&2',
        ], delay=0.2, wait=False)
        time.sleep(0.01)

        lines = []
        for line in script.stream_logs(stderr=False, tail=0, timeout=1.5):
            lines.append(line)
        self.assertEqual(lines, [b'o0\n', b'o1\n'])

    def test_stream_logs_stderr(self):
        """
        We can choose stderr when streaming logs.
        """
        script = self.run_logs_container([
            'echo "o0"', 'echo "e0" >&2',
            'echo "o1"', 'echo "e1" >&2',
        ], delay=0.2, wait=False)
        time.sleep(0.01)

        lines = []
        for line in script.stream_logs(stdout=False, timeout=1.5):
            lines.append(line)
        self.assertEqual(lines, [b'e0\n', b'e1\n'])

    def test_wait_for_logs_matching(self):
        """
        This behaves like calling the underlying wait_for_logs_matching()
        function with the container object as the first parameter.
        """
        script = self.run_logs_container([
            'echo "hi"',
            'echo "heya" >&2',
            'echo "hello"',
        ], delay=0.2, wait=False)

        script.wait_for_logs_matching(EqualsMatcher('hello'))
        with self.assertRaises(RuntimeError):
            script.wait_for_logs_matching(EqualsMatcher('goodbye'))

    def test_wait_for_logs_matching_timeout(self):
        """
        If we take too long to get a match, we time out.
        """
        script = self.run_logs_container([
            'echo "hi"',
            'echo "heya" >&2',
            'echo "hello"',
        ], delay=0.2, wait=False)
        with self.assertRaises(TimeoutError):
            script.wait_for_logs_matching(EqualsMatcher('hello'), timeout=0.1)

    def test_http_client(self):
        """
        We can get an HTTP client from the container object.
        """
        self.assertEqual(self.definition._http_clients, [])
        self.definition._create_kwargs = {
            'ports': {'8000/tcp': ('127.0.0.1', '10701')},
        }
        with self.definition as base:
            client = base.http_client()
            self.assertEqual(
                client._base_url.to_text(), 'http://127.0.0.1:10701')
            # Client is stashed for cleanup.
            self.assertEqual(self.definition._http_clients, [client])
        # Client is cleaned up at the end.
        self.assertEqual(self.definition._http_clients, [])


class TestNetworkDefinition(unittest.TestCase, DefinitionTestMixin):
    def setUp(self):
        self._setup()

    def make_definition(self, name, helper=None):
        return NetworkDefinition(name, helper=helper)


class TestVolumeDefinition(unittest.TestCase, DefinitionTestMixin):
    def setUp(self):
        self._setup()

    def make_definition(self, name, helper=None):
        return VolumeDefinition(name, helper=helper)



import unittest

from docker.models.containers import ExecResult

from seaworthy.utils import output_lines


class TestOutputLinesFunc(unittest.TestCase):
    def test_bytes(self):
        """String lines are parsed from output bytes."""
        self.assertEqual(output_lines(b'foo\nbar\n'), ['foo', 'bar'])

    def test_exec_result(self):
        """String lines are parsed from an ExecResult."""
        self.assertEqual(output_lines(ExecResult(128, b'foo\r\nbar\r\n')),
                         ['foo', 'bar'])

    def test_custom_encoding(self):
        """String lines can be parsed using a custom encoding."""
        self.assertEqual(output_lines(b'\xe1', encoding='latin1'), ['á'])



import threading
from queue import Queue


class FakeStreamSource:
    """
    A fake thing that emits canned items.

    Logs can either be streamed through a socket using ``.attach_socket()`` or
    tailed by using ``.items()``. After a log entry has been streamed, it is
    stored and we don't stream it next time. Only items that have already been
    streamed are tailed.
    """

    def __init__(self, items, expected_params=None, close_timeout=2):
        self.items = items
        self._seen_items = []
        self._expected_stream_params = {}
        if expected_params is not None:
            self._expected_stream_params.update(expected_params)
        self._feeders = set()
        self._close_timeout = close_timeout

    def cleanup(self):
        while self._feeders:
            self._feeders.pop().cancel()

    def tail_items(self, tail):
        if tail == 0:
            # Nothing to tail.
            return []
        if tail == 'all':
            # ALL THE LOGS/ITEMS!
            return self._seen_items
        # Just some of the items.
        assert tail > 0
        return self._seen_items[-tail:]

    def stream_items(self, tail, kw):
        assert kw == self._expected_stream_params
        feeder = StreamFeeder(self, self.tail_items(tail))
        self._feeders.add(feeder)
        feeder.start()
        return feeder.client_stream()


class StreamFeeder(threading.Thread):
    def __init__(self, source, tail):
        super().__init__()
        self.src = source
        self.q = Queue()
        self.tail = tail
        self.finished = threading.Event()

    def client_stream(self):
        return FakeCancellableStream(self, self.q)

    def cancel(self):
        self.finished.set()
        self.join()

    def send_item(self, item):
        self.q.put(item)

    def run(self):
        # Emit tailed items.
        for item in self.tail:
            self.send_item(item)
        # Emit previously unstreamed items at designated intervals.
        for delay, item in self.src.items[len(self.src._seen_items):]:
            # Wait for either cancelation (break) or timeout (no break).
            if self.finished.wait(delay):
                break
            self.src._seen_items.append(item)
            self.send_item(item)
        # Wait until we're done, which we may already be. Since some of the
        # tests don't do client-side timeouts, we use a fake-specific "server"
        # timeout.
        self.finished.wait(self.src._close_timeout)
        # Time to clean up.
        self.q.put(None)


class FakeCancellableStream:
    """
    Fake CancellableStream that iterates over the items.
    """

    def __init__(self, feeder, q):
        self._feeder = feeder
        self._q = q

    def __iter__(self):
        return self

    def __next__(self):
        assert self._q is not None, "Stream already closed."
        item = self._q.get()
        if item is None:
            self._q = None
            raise StopIteration
        return item

    def close(self):
        self._feeder.cancel()



import unittest
from datetime import datetime

from seaworthy.checks import docker_client, dockertest
from seaworthy.helpers import DockerHelper, fetch_images
from seaworthy.stream.logs import stream_logs, wait_for_logs_matching
from seaworthy.stream.matchers import EqualsMatcher

from .fake_stream import FakeStreamSource

# We use this image to test with because it is a small (~4MB) image from
# https://github.com/docker-library/official-images that we can run shell
# scripts in.
IMG = 'alpine:latest'


@dockertest()
def setUpModule():  # noqa: N802 (The camelCase is mandated by unittest.)
    with docker_client() as client:
        fetch_images(client, [IMG])


class FakeLogsContainer(FakeStreamSource):
    """
    A container object stub that emits canned logs.

    Logs can either be streamed through a socket using ``.attach_socket()`` or
    tailed by using ``.logs()``. After a log entry has been streamed, it is
    stored and we don't stream it next time. Only logs that have already been
    streamed are tailed.
    """

    def logs(self, stream=False, **kw):
        tail = kw.pop('tail', 'all')
        if stream:
            return self.stream_items(tail, kw)
        else:
            return b''.join(self.tail_items(tail))


class TestFakeLogsContainer(unittest.TestCase):
    def mkcontainer(self, *args, **kw):
        kw.setdefault('close_timeout', 0.1)
        con = FakeLogsContainer(*args, **kw)
        self.addCleanup(con.cleanup)
        return con

    def stream(self, con, **kw):
        return list(stream_logs(con, **kw))

    def test_empty(self):
        """
        Streaming logs for a container with no logs returns immediately.
        """
        con = self.mkcontainer([])
        self.assertEqual(self.stream(con), [])
        self.assertEqual(con.logs(tail=1), b'')

    def test_tail_only_returns_streamed(self):
        """
        Only logs that have been streamed can be tailed.
        """
        con = self.mkcontainer([(0, b'hello\n'), (0, b'goodbye\n')])
        self.assertEqual(con.logs(tail=2), b'')
        self.assertEqual(self.stream(con), [b'hello\n', b'goodbye\n'])
        self.assertEqual(con.logs(tail=2), b'hello\ngoodbye\n')

    def test_tail_tails(self):
        """
        Tailing returns the last N log lines, or all line if there are fewer
        than N or if N is 'all' (which is the default).
        """
        con = self.mkcontainer([(0, b'hello\n'), (0, b'goodbye\n')])
        self.stream(con)
        self.assertEqual(con.logs(tail=1), b'goodbye\n')
        self.assertEqual(con.logs(tail=2), b'hello\ngoodbye\n')
        self.assertEqual(con.logs(tail=3), b'hello\ngoodbye\n')
        self.assertEqual(con.logs(tail='all'), b'hello\ngoodbye\n')
        self.assertEqual(con.logs(), b'hello\ngoodbye\n')

    def test_streaming_waits(self):
        """
        Streamed logs will be returned at specified intervals. Any logs that
        have already been streamed are not returned again if we set tail=0.

        NOTE: This test measures wall-clock time, so if something causes it to
              be too slow the second assertion may fail.
        """
        con = self.mkcontainer([(0.1, b'hello\n'), (0.2, b'goodbye\n')])
        t0 = datetime.now()
        self.assertEqual(self.stream(con), [b'hello\n', b'goodbye\n'])
        t1 = datetime.now()
        self.assertEqual(self.stream(con, tail=0), [])
        t2 = datetime.now()
        self.assertLess(0.3, (t1 - t0).total_seconds())
        self.assertLess((t2 - t1).total_seconds(), 0.3)


class TestStreamLogsFunc(unittest.TestCase):
    def mkcontainer(self, *args, **kw):
        kw.setdefault('close_timeout', 0.1)
        con = FakeLogsContainer(*args, **kw)
        self.addCleanup(con.cleanup)
        return con

    def stream_only(self, con, timeout=1):
        return list(stream_logs(con, tail=0, timeout=timeout))

    def stream_logs(self, con, timeout=0.5, **kw):
        return stream_logs(con, timeout=timeout, **kw)

    def test_stream_only(self):
        """
        If there are no historical logs, we get all the streamed logs.
        """
        con = self.mkcontainer([
            (0.1, b'hello\n'),
            (0.1, b'goodbye\n'),
        ])
        self.assertEqual(
            list(self.stream_logs(con, tail=0)), [b'hello\n', b'goodbye\n'])

    def test_historical_only(self):
        """
        If all the logs have been streamed, we only get the old ones.
        """
        con = self.mkcontainer([
            (0.1, b'hello\n'),
            (0.1, b'goodbye\n'),
        ])
        self.assertEqual(self.stream_only(con), [b'hello\n', b'goodbye\n'])
        self.assertEqual(self.stream_only(con), [])
        self.assertEqual(
            list(self.stream_logs(con)), [b'hello\n', b'goodbye\n'])

    def test_timeout_and_stream_again(self):
        """
        If we take too long to get the next line, we time out. When we stream
        again, we get historical logs as well as new ones.
        """
        con = self.mkcontainer([
            (0.1, b'hello\n'),
            (0.2, b'goodbye\n'),
        ])
        lines = []
        with self.assertRaises(TimeoutError):
            for line in self.stream_logs(con, timeout=0.15):
                lines.append(line)
        self.assertEqual(lines, [b'hello\n'])
        lines = list(self.stream_logs(con, timeout=0.35))
        self.assertEqual(lines, [b'hello\n', b'goodbye\n'])


class TestWaitForLogsMatchingFunc(unittest.TestCase):
    def mkcontainer(self, *args, **kw):
        kw.setdefault('close_timeout', 0.1)
        con = FakeLogsContainer(*args, **kw)
        self.addCleanup(con.cleanup)
        return con

    def wflm(self, con, matcher, timeout=0.5, **kw):
        return wait_for_logs_matching(con, matcher, timeout=timeout, **kw)

    def test_one_matching_line(self):
        """
        If one matching line is logged, all is happy.
        """
        con = self.mkcontainer([
            (0, b'hello\n'),
        ])
        # If this doesn't raise an exception, the test passes.
        self.wflm(con, EqualsMatcher('hello'))

    def test_one_non_matching_line(self):
        """
        If there's no match by the time the logs end, we raise an exception.
        """
        con = self.mkcontainer([
            (0, b'goodbye\n'),
        ])
        with self.assertRaises(RuntimeError) as cm:
            self.wflm(con, EqualsMatcher('hello'))
        self.assertIn(
            "Logs matching EqualsMatcher('hello') not found.",
            str(cm.exception))
        self.assertIn('goodbye\n', str(cm.exception))

    def test_timeout_first_line(self):
        """
        If we take too long to get the first line, we time out.
        """
        con = self.mkcontainer([
            (0.2, b'hello\n'),
        ])
        with self.assertRaises(TimeoutError) as cm:
            self.wflm(con, EqualsMatcher('hello'), timeout=0.1)
        self.assertIn(
            "Timeout (0.1s) waiting for logs matching EqualsMatcher('hello').",
            str(cm.exception))
        self.assertNotIn('hello\n', str(cm.exception))

    def test_timeout_later_line(self):
        """
        If we take too long to get a later line, we time out.
        """
        con = self.mkcontainer([
            (0, b'hi\n'),
            (0.2, b'hello\n'),
        ])
        with self.assertRaises(TimeoutError) as cm:
            self.wflm(con, EqualsMatcher('hello'), timeout=0.1)
        self.assertIn(
            "Timeout (0.1s) waiting for logs matching EqualsMatcher('hello').",
            str(cm.exception))
        self.assertIn('hi\n', str(cm.exception))
        self.assertNotIn('hello\n', str(cm.exception))

    def test_default_encoding(self):
        """
        By default, we assume logs are UTF-8.
        """
        con = self.mkcontainer([
            (0, b'\xc3\xbeorn\n'),
        ])
        # If this doesn't raise an exception, the test passes.
        self.wflm(con, EqualsMatcher('\u00feorn'))

    def test_encoding(self):
        """
        We can operate on logs that use excitingly horrible encodings.
        """
        con = self.mkcontainer([
            (0, b'\xfeorn\n'),
        ])
        # If this doesn't raise an exception, the test passes.
        self.wflm(con, EqualsMatcher('\u00feorn'), encoding='latin1')

    def test_kwargs(self):
        """
        We pass through any kwargs we don't recognise to docker.
        """
        con = self.mkcontainer([(0, b'hi\n')], {'stdout': False})
        with self.assertRaises(AssertionError):
            self.wflm(con, EqualsMatcher('hi'))
        with self.assertRaises(AssertionError):
            self.wflm(con, EqualsMatcher('hi'), stdout=False, stderr=False)
        self.wflm(con, EqualsMatcher('hi'), stdout=False)


class FakeAndRealContainerMixin:
    def wflm(self, con, matcher, timeout=0.5, **kw):
        return wait_for_logs_matching(con, matcher, timeout=timeout, **kw)

    def stream_logs(self, con, timeout=0.5, **kw):
        for line in stream_logs(con, timeout=timeout, **kw):
            yield line

    def test_fake_and_real_logging_behaviour(self):
        """
        Our fake logs container should exhibit similar behaviour to a real
        container.
        """
        logger = self.start_logging_container()
        self.wflm(logger, EqualsMatcher('Log entry 1'), timeout=1)

        early_logs = logger.logs().decode('utf8').splitlines()
        self.assertIn('Log entry 1', early_logs)
        self.assertNotIn('Log entry 4', early_logs)

        streamed_logs = []
        with self.assertRaises(TimeoutError):
            for line in self.stream_logs(logger, tail=0, timeout=0.7):
                streamed_logs.append(line.decode('utf8').rstrip())
        self.assertNotIn('Log entry 1', streamed_logs)
        self.assertIn('Log entry 4', streamed_logs)
        self.assertNotIn('Log entry 9', streamed_logs)

        streamed_logs = []
        for line in self.stream_logs(logger, timeout=2):
            streamed_logs.append(line.decode('utf8').rstrip())
        self.assertIn('Log entry 1', streamed_logs)
        self.assertIn('Log entry 4', streamed_logs)
        self.assertIn('Log entry 9', streamed_logs)


class TestWithFakeContainer(unittest.TestCase, FakeAndRealContainerMixin):
    def mkcontainer(self, *args, **kw):
        kw.setdefault('close_timeout', 0.1)
        con = FakeLogsContainer(*args, **kw)
        self.addCleanup(con.cleanup)
        return con

    def start_logging_container(self):
        return self.mkcontainer([
            (0.2, 'Log entry {}\n'.format(n).encode('utf8'))
            for n in range(10)])


@dockertest()
class TestWithRealContainer(unittest.TestCase, FakeAndRealContainerMixin):
    def setUp(self):
        self.dh = DockerHelper()
        self.addCleanup(self.dh.teardown)

    def start_logging_container(self):
        script = '\n'.join([
            'for n in 0 1 2 3 4 5 6 7 8 9; do',
            '    sleep 0.2',
            '    echo "Log entry ${n}"',
            'done',
        ])
        logger = self.dh.containers.create(
            'logger', IMG, command=['sh', '-c', script])
        self.addCleanup(self.dh.containers.remove, logger, force=True)
        logger.start()
        logger.reload()
        return logger



import unittest

from seaworthy.stream.matchers import (
    EqualsMatcher, OrderedMatcher, RegexMatcher, UnorderedMatcher)


class TestEqualsMatcher(unittest.TestCase):
    def test_matching(self):
        """ Matches exactly equal strings and nothing else. """
        matcher = EqualsMatcher('foo')
        self.assertTrue(matcher('foo'))
        self.assertFalse(matcher('foobar'))

    def test_str(self):
        """ The string representation is readable. """
        matcher = EqualsMatcher('bar')
        self.assertEqual(str(matcher), "EqualsMatcher('bar')")
        self.assertEqual(repr(matcher), str(matcher))


class TestRegexMatcher(unittest.TestCase):
    def test_matching(self):
        """
        Matches strings that match the pattern, doesn't match other strings.
        """
        matcher = RegexMatcher(r'^foo')
        self.assertTrue(matcher('foobar'))
        self.assertFalse(matcher('barfoo'))

    def test_str(self):
        """ The string representation is readable. """
        matcher = RegexMatcher(r'^bar')
        self.assertEqual(str(matcher), "RegexMatcher('^bar')")
        self.assertEqual(repr(matcher), str(matcher))


class TestOrderedMatcher(unittest.TestCase):
    def test_matching(self):
        """
        Applies each matcher sequentially and returns True on the final match.
        """
        matcher = OrderedMatcher(
            EqualsMatcher('foo'), RegexMatcher('^bar'))

        self.assertFalse(matcher('barfoo'))
        self.assertFalse(matcher('foo'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('barfoo'))

    def test_by_equality(self):
        """
        The ``by_equality`` utility method takes a list of strings and produces
        a matcher that matches those strings by equality sequentially.
        """
        matcher = OrderedMatcher.by_equality('foo', 'bar')

        self.assertFalse(matcher('bar'))
        self.assertFalse(matcher('foo'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('bar'))

    def test_by_regex(self):
        """
        The ``by_regex`` utility method takes a list of patterns and produces a
        matcher that matches by those regex patterns sequentially.
        """
        matcher = OrderedMatcher.by_regex(r'^foo', r'bar$')

        self.assertFalse(matcher('fuzzbar'))
        self.assertFalse(matcher('foobar'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('foobar'))

    def test_exhaustion(self):
        """
        Once all matchers have been matched, further calls to ``match`` should
        raise an error.
        """
        matcher = OrderedMatcher.by_equality('foo')
        self.assertTrue(matcher('foo'))

        with self.assertRaises(RuntimeError) as cm:
            matcher('bar')
        self.assertEqual(str(cm.exception),
                         'Matcher exhausted, no more matchers to use')

    def test_str(self):
        """
        The string representation is readable and shows which matchers have
        been matched and which are still to be matched.
        """
        matcher = OrderedMatcher(
            EqualsMatcher('foo'), RegexMatcher('^bar'))

        self.assertEqual(
            str(matcher),
            'OrderedMatcher(matched=[], '
            "unmatched=[EqualsMatcher('foo'), RegexMatcher('^bar')])")
        self.assertEqual(repr(matcher), str(matcher))

        self.assertFalse(matcher('foo'))

        self.assertEqual(
            str(matcher),
            "OrderedMatcher(matched=[EqualsMatcher('foo')], "
            "unmatched=[RegexMatcher('^bar')])")
        self.assertEqual(repr(matcher), str(matcher))

        self.assertTrue(matcher('barfoo'))

        self.assertEqual(
            str(matcher),
            "OrderedMatcher(matched=[EqualsMatcher('foo'), "
            "RegexMatcher('^bar')], unmatched=[])")
        self.assertEqual(repr(matcher), str(matcher))


class TestUnorderedMatcher(unittest.TestCase):
    def test_matching(self):
        """
        Applies each matcher sequentially and returns True on the final match.
        """
        matcher = UnorderedMatcher(
            EqualsMatcher('foo'), RegexMatcher('^bar'))

        self.assertFalse(matcher('barfoo'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('foo'))

        matcher = UnorderedMatcher(
            EqualsMatcher('foo'), RegexMatcher('^bar'))

        self.assertFalse(matcher('foo'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('barfoo'))

    def test_by_equality(self):
        """
        The ``by_equality`` utility method takes a list of strings and produces
        a matcher that matches those strings by equality sequentially.
        """
        matcher = UnorderedMatcher.by_equality('foo', 'bar')

        self.assertFalse(matcher('foo'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('bar'))

    def test_by_regex(self):
        """
        The ``by_regex`` utility method takes a list of patterns and produces a
        matcher that matches by those regex patterns sequentially.
        """
        matcher = UnorderedMatcher.by_regex(r'^foo', r'bar$')

        self.assertFalse(matcher('foobar'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('foobar'))

    def test_exhaustion(self):
        """
        Once all matchers have been matched, further calls to ``match`` should
        raise an error.
        """
        matcher = UnorderedMatcher.by_equality('foo')
        self.assertTrue(matcher('foo'))

        with self.assertRaises(RuntimeError) as cm:
            matcher('bar')
        self.assertEqual(str(cm.exception),
                         'Matcher exhausted, no more matchers to use')

    def test_str(self):
        """
        The string representation is readable and shows which matchers have
        been matched and which are still to be matched.
        """
        matcher = UnorderedMatcher(
            EqualsMatcher('foo'), RegexMatcher('^bar'))

        self.assertEqual(
            str(matcher),
            'UnorderedMatcher(matched=[], '
            "unmatched=[EqualsMatcher('foo'), RegexMatcher('^bar')])")
        self.assertEqual(repr(matcher), str(matcher))

        self.assertFalse(matcher('foo'))

        self.assertEqual(
            str(matcher),
            "UnorderedMatcher(matched=[EqualsMatcher('foo')], "
            "unmatched=[RegexMatcher('^bar')])")
        self.assertEqual(repr(matcher), str(matcher))

        self.assertTrue(matcher('barfoo'))

        self.assertEqual(
            str(matcher),
            "UnorderedMatcher(matched=[EqualsMatcher('foo'), "
            "RegexMatcher('^bar')], unmatched=[])")
        self.assertEqual(repr(matcher), str(matcher))
